package edu.toronto.cs.algebra;

import edu.toronto.cs.util.*;
import org.w3c.dom.*;
import java.util.*;

import java.io.File;

/*** constructs a lattice */
public class LatticeFactory
{
  // -- create a logger
  static Logger log = new Logger (LatticeFactory.class.getName ());
  
  // -- tester method
  public static void main (String[] args) throws Exception
  {
    FiniteLattice fLattice = parse (args [0]);
    System.out.println ("Parsed a lattice:");
    System.out.println ("Dumping tables");
    java.io.PrintWriter out = new java.io.PrintWriter (System.out);
    out.println ("HELLO");
    fLattice.dumpTables (out);
    out.println ("Carrier set: ");
    for (IntIterator it = fLattice.carrierSet (); it.hasNext ();)
      out.println (it.nextInt ());
    out.flush ();
  }
  

  public static FiniteLattice parse (String uri) throws Exception
  {
    System.out.println ("Parsing: " + uri);
    // -- first parse the XML document
    Document document = DOMUtil.parse (uri);
    // -- construct a QBLattice from XML Document
    return buildLattice (document);
  }
  

  public static FiniteLattice buildLattice (Document document)
    throws Exception
  {
    NodeList graphs = document.getElementsByTagName ("graph");
    if (graphs == null || graphs.getLength () != 1)
      throw new RuntimeException 
	("XML Document should contain one graph node");
    
    FiniteLattice lattice = buildLattice ((Element)graphs.item (0));
    // -- set other lattice arguments, like image and things
    
    if (document.getElementsByTagNameNS ("*", "img").getLength () != 0)
      {
	Element imgElement = 
	  (Element)document.getElementsByTagNameNS ("*", "img").item (0);
	String imagePath = imgElement.getAttribute ("xlink:href");
	log.debug ("Logic image is: " + imagePath);
	File imageFile = new File (imagePath);
	if (imageFile.exists ())
	  {
	    log.debug ("Image file found");
	    lattice.setImageFile (imageFile);
	  }
	else
	  log.debug ("File not found: " + imageFile);
      }
    
    return lattice;
  }
  
  /***
   * Given a top level graph node constructs a FiniteLattice
   */
  public static FiniteLattice buildLattice (Element graph) throws Exception
  {
    if (!sanityCheck (graph)) return null;

    // -- to construct a FiniteLattice we need to know how many elements
    // -- we have in the lattice (= # of node tags)
    // -- and convert the 'above' edges into a list
    // -- and convert the 'neg' edges into a list
    int latticeSize = graph.getElementsByTagName ("node").getLength ();

    List elements = collectElementNames (graph.getElementsByTagName ("node"));
    List order = buildOrder (graph.getElementsByTagName ("edge"));
    List negation = buildNegation (graph.getElementsByTagName ("edge"));
    
    return newFiniteLattice (elements, order, negation);
  }
  
  public static FiniteLattice newFiniteLattice (List elements, List lorder, 
						List lnegation)
    throws Exception
  {
    return new FiniteLattice (elements, lorder, lnegation);
  }

  
  static List buildOrder (NodeList edges)
  {
    List order = new ArrayList ();
    
    int size = edges.getLength ();
    for (int i = 0; i < size; i++)
      {
	Element edge = (Element)edges.item (i);
	Element type = (Element)edge.getElementsByTagName ("type").item (0);
	if (type.getAttribute ("value").equals ("above"))
	  {
	    Object[] entry = new Object [2];
	    entry [0] = edge.getAttribute ("from");
	    entry [1] = edge.getAttribute ("to");
	    order.add (entry);
	  }
      }
    return order;
  }

  static List buildNegation (NodeList edges)
  {
    List negation = new ArrayList ();
    
    int size = edges.getLength ();
    for (int i = 0; i < size; i++)
      {
	Element edge = (Element)edges.item (i);
	Element type = (Element)edge.getElementsByTagName ("type").item (0);
	if (type.getAttribute ("value").equals ("neg"))
	  {
	    Object[] entry = new Object [2];
	    entry [0] = edge.getAttribute ("from");
	    entry [1] = edge.getAttribute ("to");
	    negation.add (entry);
	  }
      }
    return negation;
  }


  static List collectElementNames (NodeList nodes)
  {
    int size = nodes.getLength ();
    List names = new ArrayList ();
    
    for (int i = 0; i < size; i++)
      {
	Element e = (Element)nodes.item (i);
	names.add (e.getAttribute ("ID"));
      }
    return names;
  }
  
  
  private static boolean sanityCheck (Element graph)
  {
    // -- no check at the moment
    return true;
  }
  
  

}
