package edu.toronto.cs.ctl;

import edu.toronto.cs.algebra.*;
import java.util.*;

/**
 ** This CTL simplifier treats !, /\, and \/ as well as EXf, E[fUp],
 ** and EGf as the adequate set of operations.
 **/
public class CTLEXEUEGSimplifier implements CTLSimplifier
{

  /**
   ** Simplifies CTL formulas. If the formula cannot be simplified for
   ** any reasons, the formula is returned unchanged.
   **
   ** @param ctl -- CTL formula to be considered.
   **
   ** @param lattice -- the lattice to be used.
   **
   ** @return -- simplified CTL.
   **/
  public CTLNode simplify (CTLNode ctl, FiniteLattice lattice)
  {
    if (ctl instanceof CTLImplNode)
      return simplifyImpl ((CTLImplNode) ctl);
    else if (ctl instanceof CTLIffNode)
      return simplifyIff ((CTLIffNode) ctl);
    else if (ctl instanceof CTLAXNode)
      return simplifyAX ((CTLAXNode) ctl);
    else if (ctl instanceof CTLEUiNode)
      return simplifyEUi ((CTLEUiNode) ctl);
    //  	else if (ctl instanceof CTLAUiNode)
    //  	  return simplifyAUi ((CTLAUiNode) ctl, lattice);
    else if (ctl instanceof CTLAUNode)
      return simplifyAU ((CTLAUNode) ctl);
    else if (ctl instanceof CTLEFNode)
      return simplifyEF ((CTLEFNode) ctl, lattice);
    else if (ctl instanceof CTLAFNode)
      return simplifyAF ((CTLAFNode) ctl, lattice);
    else if (ctl instanceof CTLAGNode)
      return simplifyAG ((CTLAGNode) ctl);
    else if (ctl instanceof CTLAWNode)
      return simplifyAW ((CTLAWNode) ctl);
    else if (ctl instanceof CTLEWNode)
      return simplifyEW ((CTLEWNode) ctl);
    else
      return ctl;
  }

  // -- all the newctl = explanations

  CTLNode simplifyEW (CTLEWNode formula)
  {
    // E (p W s) = E(p U s) | EG p
    CTLNode p = formula.getLeft ();
    CTLNode s = formula.getRight ();
    CTLNode epus = CTLFactory.createCTLEUNode (p, s, formula.getFairness ());
    CTLNode egp = CTLFactory.createCTLEGNode (p, formula.getFairness ());
   
    return epus.or (egp);
  }
  

  CTLNode simplifyAW (CTLAWNode formula)
  {
    // A (p W s) = !E (!s U !(s | p))
    CTLNode p = formula.getLeft ();
    CTLNode s = formula.getRight ();
    CTLNode eu = CTLFactory.createCTLEUNode (s.neg (),
					     (s.or (p)).neg (),
					     formula.getFairness ());

    return eu.neg ();
  }
  
  CTLNode simplifyImpl (CTLImplNode formula)
  {
    // -- use a -> b <=> !a | b
    CTLNode a = formula.getLeft ();
    CTLNode b = formula.getRight ();
    return a.neg ().or (b);
  }

  CTLNode simplifyIff (CTLIffNode formula)
  {
    // -- use a <-> b <=> a -> b & a <- b
    CTLNode a = formula.getLeft ();
    CTLNode b = formula.getRight ();
    return a.implies (b).and (a.impliedBy (b));
  }

  CTLNode simplifyAX (CTLAXNode formula)
  {
    // -- use AX f <=> !EX!f
    CTLNode f = formula.getRight ();
    CTLNode exnf = CTLFactory.createCTLEXNode (f.neg (),
					       formula.getFairness ());
    return exnf.neg ();
  }

  CTLNode simplifyAF (CTLAFNode formula, FiniteLattice lat)
  {
    // -- use AF f <=> A[true U f]
    CTLNode f = formula.getRight ();

    CTLLatticeNode top = CTLFactory.createCTLLatticeNode (lat.idtimesIdx (),
							  lat);
    CTLNode atuf = CTLFactory.createCTLAUNode (top, f,
					       formula.getFairness ());
    return atuf;
  }

  CTLNode simplifyEF (CTLEFNode formula, FiniteLattice lat)
  {
    // -- use EF f <=> E[true U f]
    CTLNode f = formula.getRight ();

    CTLLatticeNode top = CTLFactory.createCTLLatticeNode (lat.idtimesIdx (),
							  lat);
    CTLNode etuf = CTLFactory.createCTLEUNode (top, f,
					       formula.getFairness ());
    return etuf;
  }

  CTLNode simplifyAG (CTLAGNode formula)
  {
    // -- use AG f <=> !EF!f
    CTLNode f = formula.getRight ();
    CTLNode efnf = CTLFactory.createCTLEFNode (f.neg (),
					       formula.getFairness ());
    return efnf.neg ();
  }

  CTLNode simplifyAU (CTLAUNode formula)
  {
    // -- use A[p U q] = !(E[!q U !(q | p)] | EG!q)
    CTLNode p = formula.getLeft ();
    CTLNode q = formula.getRight ();
    CTLNode eu = CTLFactory.createCTLEUNode (q.neg (),
					     (q.or (p)).neg (),
					     formula.getFairness ());
    CTLNode eg = CTLFactory.createCTLEGNode (q.neg (),
					     formula.getFairness ());

    return (eu.or (eg)).neg ();
  }
  /*
  CTLNode simplifyAUi (CTLAUiNode formula)
  {
    // -- use A[f Ui p] <=> p | f & AX A[f ui-1 p] & EX A[f Ui-1 p]
    CTLNode f = formula.getLeft ();
    CTLNode p = formula.getRight ();
    int i = formula.getI();
    if (i == 0)
      return p;
    else
      {
	CTLAUiNode afup = f.au (i-1, p);
	return p.or (f.and (afup.ax ().and (afup.ex ())));
      }
  }
  */
  CTLNode simplifyEUi (CTLEUiNode formula)
  {
    // -- use E[f Ui p] <=> p | f & EX E[f ui-1 p]
    CTLNode f = formula.getLeft ();
    CTLNode p = formula.getRight ();
    int i = formula.getI();

    if (i == 0)
      return p;
    else
      {
	CTLEUiNode efup = f.eu (i-1, p);
	return p.or (f.and (efup.ex ()));
      }
  }
}
