package edu.toronto.cs.keg;

import java.util.*;
import java.io.*;

import edu.toronto.cs.algebra.*;
import edu.toronto.cs.util.dgraph.*;
import edu.toronto.cs.util.*;
import edu.toronto.cs.ctl.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.mdd.*;
import edu.toronto.cs.heuristics.*;
import edu.toronto.cs.modelchecker.*;

/**
 ** This is the module that generates new states for counter-examples
 ** using MDDs.
 **/
public class KEMDDStateGenerator implements KEGExpander
{
  protected HeuristicAnalyser heuristics;
  
  /**
   ** Get the heuristics.
   ** @return value of heuristics.
   **/
  public HeuristicAnalyser getHeuristics() {
    return heuristics;
  }
  
  /**
   ** Set the heuristics.
   ** @param h -- heuristics to be used.
   **/
  public void setHeuristics(HeuristicAnalyser h) {
    heuristics = h;
  }

  /**
   ** Constructs the simplifier without heuristics.
   **/
  public KEMDDStateGenerator ()
  {
    heuristics = new NoHeuristics ();
  }

  /**
   ** Constructs the simplifier with specifed heuristics.
   **/
  public KEMDDStateGenerator (HeuristicAnalyser h)
  {
    heuristics = h;
  }

  /**
   ** Attempts to expand a given state in order to provide a
   ** counter-example.
   **
   ** @param state -- a node in the state graph that contains the state
   ** to be considered.
   **
   ** @param formula -- a node in the CTL graph that contains the
   ** formula to be considered.
   **
   ** @param mc -- the model checker to be used.
   **
   ** @return -- true if the expansion was successful; false otherwise.
   **/
  public boolean expand (DGNode state, DGNode formula, MvSetModelChecker mc)
  {

    KEGCtlNode ctlnode = (KEGCtlNode) formula;
    DGraph ctlgraph = ctlnode.getGraph ();

    KEGStateNodeFactory factory = (KEGStateNodeFactory) state.getFactory ();
    MvState s0 = ((KEGStateNode) state).getState ();

    
    // -- break the CTL into formula + value
    CTLNode ctl = ctlnode.getCTL ().getLeft ();
    CTLLatticeNode value = (CTLLatticeNode) ctlnode.getCTL ().getRight ();
    // -- the new states that we add
    Set newstates;

    /* System.out.println ("Expand with lattice:");
    PrintWriter out = new PrintWriter (System.out);    
    mc.getLattice ().dumpValueMap (out);
    out.flush (); */
    // check whether the formula involves EX or not
    if (ctl instanceof CTLEXNode)
      {
	//	System.out.println ("Justifying an EX with a state");
	//	System.out.println (mc.getLattice ().idplusIdx ());
	// -- make sure that the value of EX phi is not BOTTOM
	if (value.getLatIdx () != mc.getLattice ().idplusIdx ())
	  {
	    //  System.out.println ("   Explanation possible.");
	    
	    newstates = explainEX ((CTLEXNode) ctl, value, s0, mc);
	    Assert.assert (!newstates.isEmpty (), "explainEX failed");
	    System.out.println ("Explaining by states: " + newstates);
	    
	    // -- now add all the new states to the graph
	    
	    Iterator i = newstates.iterator ();
	    int trans;
	    MvState succ;
	    KEGStateNode nextState;
	    System.out.println ("Maketerm.");
	    //((KEGCtlNode)ctlnode).makeTerminal (newstates);
	    while (i.hasNext ())
	      {
		// -- get the successor state
		succ = (MvState) i.next ();
		// -- get the transition value

		trans = mc.getMvTransRelation ().restrict (s0)
		  .renameArgs (mc.getPropOrdering ().getUnPrimedMap ())
		  .evaluate (succ);
		KegUtils.addState 
		  (state, formula, succ, trans, ctl.getRight ());
		/*
		nextState = factory.createNode ((KEGStateNode)state, succ);
		ctlgraph = new DGraph (new KEGCtlNodeFactory ());
		ctlgraph.addRootNode (((KEGCtlNodeFactory)
				       ctlgraph.getFactory ()).createNode 
				      (ctl.getRight ()));
		nextState.setCTLGraph (ctlgraph);
		((KEGStateNode) state).add (nextState, trans);
		*/

	      }
 
	    return true;
	  }
	else // -- can't give an explanation for EX phi = BOTTOM
	  {
	    System.out.println 
	      ("Cannot exlain EX phi=Bottom: " + ctlnode.getCTL ());
	    return false;
	  }
	
      }
    else
      return false;
  }

  // -- all the methods used for expansion of the states

  // -- produces the set of states that show why the value of EX phi
  // -- is not any lower than it is
  Set explainEX (CTLEXNode ctl, CTLLatticeNode value, MvState s0,
		 MvSetModelChecker mc)
  {
    Set states;
    CTLNode rightCTL = ctl.getRight ();
    // -- get the MvSet for R/\phi'
    MvSet preEX = mc.checkCTL (CTLFactory.createCTLPreEXNode (rightCTL));

    preEX = preEX.restrict (s0);
    // -- unprime
    preEX = preEX.renameArgs (mc.getPropOrdering ().getUnPrimedMap ());
    // -- get the MDD for the MvSet to perform some smart things on it...
    MDDNode mddNode = ((MDDMvSet) preEX).getMddNode ();
    // -- get the important states
    System.out.println ("We are looking for the interesting values out of:");
    System.out.println (mddNode.getImage ());
    BitSet vals = mc.getLattice ().getIrredundantJoin (mddNode.getImage ());
    System.out.println ("Which is:");
    System.out.println (vals);
    
    /*** check if this is a splitting (branching) point ***/
    List setBits = getOnBits (vals);
    if (setBits.size () > 1 && rightCTL.getClass () == CTLEUiNode.class &&
	((CTLEUiNode)rightCTL).getI () > 0)
      return explainEXEUi (ctl, s0, mc, setBits);
    
    states = mddNode.findStates (vals);
    return states;
  }

  List getOnBits (BitSet vals)
  {
    List list = new ArrayList ();
    for (int i = 0; i < vals.size (); i++)
      if (vals.get (i)) list.add (new Integer (i));
    return list;
  }
  

  Set explainEXEUi (CTLEXNode ctl,  
		    MvState s0, MvSetModelChecker mc, List values)
  {

    System.out.println ("Explaining: " + ctl);
    /***
     ** We need to explain EXE[\varphi U_i \varpsi] 
     ** by several paths. However, we want each path to be minimal
     ** and so we proceed to search for the smallest i
     ***/

    // -- stores all of the states we find
    Set states = new HashSet ();

    CTLEUiNode eui = (CTLEUiNode)ctl.getRight ();
    CTLNode phi = eui.getLeft ();
    CTLNode psi = eui.getRight ();
    
    int bound = eui.getI ();
    
    Algebra lattice = mc.getLattice ();
    
    int prevValue = FiniteLattice.NO_VALUE;
    
    for (int i = bound - 1; i >= 0 && !values.isEmpty (); i--)
      {
	System.out.println ("EXEUi: bound is set at: " + i);
	BitSet valuesToFind = new BitSet ();
	// -- find the value of EXE[phi U_bound psi] at s0
	int value = mc.checkCTL (phi.eu (i, psi).ex ()).evaluate (s0);
	System.out.println ("EXEUi: checking: " + 
			    phi.eu (i, psi).ex ());
	
	System.out.println ("EXEUi: model checking to get: " + value);
	if (value == prevValue) continue;
	
	prevValue = value;
	
	for (Iterator it = values.iterator (); it.hasNext ();)
	  {
	    int euiVal = ((Integer)it.next ()).intValue ();
	    if (!lattice.leq (euiVal, value))
	      {
		System.out.println ("EXEUi: adding " + euiVal + 
				    " because of " + value);
		valuesToFind.set (euiVal);
		it.remove ();
	      }
	  }

	System.out.println ("EXEUi iteration, got " + 
			    getOnBits (valuesToFind).size ()
			    + " values to explain");
	if (valuesToFind.length () > 0)
	  {
	    MvSet preEX = 
	      mc.checkCTL 
	      (CTLFactory.createCTLPreEXNode (phi.eu (i + 1, psi))).
	      restrict (s0).
	      renameArgs (mc.getPropOrdering ().getUnPrimedMap ());
	    MDDNode mddNode = ((MDDMvSet)preEX).getMddNode ();
	    
	    states.addAll (mddNode.findStates (valuesToFind));
	  }
      }
    

    System.out.println ("Doing EXEUi expansion with " + states.size () + 
			" states");
    return states;
  }
  
  
}
