package edu.toronto.cs.lift;

import java.util.*;

import edu.toronto.cs.smv.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.util.*;


// XXX for testing only
import edu.toronto.cs.modelchecker.AlternatePropOrder;
import edu.toronto.cs.algebra.*;

/****
 **** LiftR with the 2/3-full feature
 ***  see tt-full.feat
 ***/

public class LiftRTT extends SMVCombinedModule
{
  ButtonR[] liftBtns;
  int floors;

  int topFloor;
  int botFloor = 1;
  
  static final int DOOR = 0;
  static final int DIR = 1;
  static final int TTFULL = 2;
  
  int floorBaseIdx;
  int floorWidth;

  MvSet noCall;
  
  // -- XXX This is needed to reference the main module due to some
  // -- XXX technical problems which should be solved eventually
  MainLiftRTT main;
  
  public LiftRTT (String _name, int _floors, MainLiftRTT _main)
  {
    super (_name);
    floors = _floors;
    topFloor = floors;

    floorBaseIdx = 3;
    floorWidth = computeFloorWidth (topFloor);
    
    vars = new String [3 + floorWidth];
    
    vars [DOOR] = "door";
    vars [DIR] = "dir";
    vars [TTFULL] = "tt_full";
    
    for (int i = floorBaseIdx; i < floorBaseIdx + floorWidth; i++)
      vars [i] = "floor" + (i - floorBaseIdx);
    

    // -- do the buttons
    liftBtns = new ButtonR [floors];
    for (int i = 1; i <= liftBtns.length; i++)
      liftBtns [i-1] = new ButtonR (getName () + 
				    SMVModule.VAR_SEP + "liftBut" + i);
    addSubModules (Arrays.asList (liftBtns));

    main = _main;
  }


  public void setNoCall (MvSet v)
  {
    noCall = v;
  }
  
  
  // -- compute number of bits required to represent v
  int computeFloorWidth (int v)
  {
    int width = 1;

    while ((v >> width) > 0) width++;
    return width;
  }
  

  public MvSet computeTrans ()
  {
    System.out.println ("Computing trans for " + getName ());

    MvSet result = computeDir ().and  
      (computeDoor ().and 
       (computeButtons ().and 
	(computeFloor ())));    

    result = result.and (computeTTFull ());
    return result;
  }
  
  MvSet computeTTFull ()
  {
    /***
	next (tt_full) := case
	                 door = closed: tt_full;
			 1: {true, false};
			 esac;
    */

    MvSet ttFullIsTTFull = 
      ddManager.createProjection (getNextVarOrder (TTFULL)). 
      eq (ddManager.createProjection (getVarOrder (TTFULL)));
    
    
    return new CaseStatement ().
      addCase (cur (DOOR).not (), ttFullIsTTFull).
      addDefault (next (TTFULL, top).or (next (TTFULL, bot))).compute ();
  }
  
  MvSet computeButtons ()
  {
    MvSet result = ddManager.createConstant (lattice.idtimesIdx ());
    
    MvSet doorOpen = cur (DOOR, top);
    for (int i = 0; i < liftBtns.length; i++)
      {
	liftBtns [i].setReset (computeFloorEquals (i+1).and (doorOpen));
	result = result.and (liftBtns [i].computeTrans ());
	if (i == 0 && result instanceof MDDMvSet)
	  ((MDDMvSet)result).writeToDaVinci ("ButtonR.daVinci", ordering);
      }

    return result;
  }

  
  MvSet computeDir ()
  {
    /*
    ** up is true, down is false **
    next(dir) := case
               idle : dir;
	       floor = topFloor : false;
	       floor = botFloor : true;
	       car_call = 0 & landing_call = 0 & dir = false : true;
	       car_call = 0 & landing_call = 0 & dir = true : false;
	       1  : dir;
	       esac;
    */

    // -- mvset for next(dir) = dir
    MvSet dirIsDir = ddManager.createProjection (getNextVarOrder (DIR));
    dirIsDir = dirIsDir.eq (ddManager.createProjection (getVarOrder (DIR)));

    // -- next (dir) = up
    MvSet dirIsUp = next (DIR, top);
    // -- next (dir) = down
    MvSet dirIsDown = next (DIR, bot);

    return new CaseStatement ().
      addCase (computeIdle (), dirIsDir).
      addCase (computeFloorEquals (topFloor), dirIsDown).
      addCase (computeFloorEquals (botFloor), dirIsUp).
      addCase (cur (DIR, bot).and (computeCarCallEquals (0)).
	       and (computeLandingCallEquals (0)), 
			dirIsUp).
      addCase (cur (DIR, top).and (computeCarCallEquals (0)).
	       and (computeLandingCallEquals (0)),
			dirIsDown).
      addDefault (dirIsDir).compute ();
  }
  
  MvSet computeFloor ()
  {
    /***
	next (floor) := case
	               door = true : floor;
		       car_call = 0 & landing_call = 0: floor;
		       direction = true & floor < 5 : floor + 1;
		       direction = flase & floor > 1 : floor - 1;
		       1: floor;
		       esac;
    */
     
    MvSet floorIsFloor = computeFloorIsFloor ();
    MvSet floorIsFloorPlusOne = computeFloorIsFloorPlusOne ();
    MvSet floorIsFloorMinusOne = computeFloorIsFloorMinusOne ();
    
    MvSet doorIsOpen = cur (DOOR, top);
    MvSet dirIsDown = cur (DIR, bot);
    MvSet dirIsUp = cur (DIR, top);
    
    return new CaseStatement ().
      addCase (doorIsOpen, floorIsFloor).
      addCase (computeCarCallEquals (0).and 
	       (computeLandingCallEquals (0)), floorIsFloor).
      addCase (dirIsUp.and (computeFloorLess (topFloor)), 
	       floorIsFloorPlusOne).
      addCase (dirIsDown.and (computeFloorGreater (botFloor)), 
	       floorIsFloorMinusOne).
      addDefault (floorIsFloor).
      compute ();
  }

  // -- compute 'floor = v'
  MvSet computeFloorEquals (int v)
  {
    return computeNatEquals (floorBaseIdx, floorWidth, v);
  }
  
  // -- computes 'floor < v'
  MvSet computeFloorLess (int v)
  {
    MvSet result = ddManager.createConstant (bot);
    
    for (int i = botFloor; i < v; i++)
      result = result.or (computeNatEquals (floorBaseIdx, floorWidth, i));
    return result;
  }
  
  MvSet computeFloorGreater (int v)
  {
    MvSet result = ddManager.createConstant (bot);
    
    for (int i = topFloor; i > v; i--)
      result = result.or (computeNatEquals (floorBaseIdx, floorWidth, i));
    return result;
  }
  
  MvSet computeDoor ()
  {
    /**
       door := case
                floor = car_call : true;
		floor = landing_call : true;
                1 : false;
               esac
    */

    MvSet curCase = new CaseStatement ().
      addCase (computeFloorIsCarCall (), cur (DOOR, top)).
      addCase (computeFloorIsLandingCall (), cur (DOOR, top)).
      addDefault (cur (DOOR, bot)).compute ();

    // -- we also need to compute the next case for the door, 
    // -- that is
    /* next(door) := case
         next(floor) = next (car_call) : true;
	 1 : false;
	 esac;
    **/
    // -- we acheive this by shifting the variables in the resulting mvset
    curCase = curCase.and (curCase.renameArgs (ordering.getPrimedMap ()));
    return curCase;
    
  }
  
  MvSet computeFloorIsCarCall ()
  {
    return computeFloorIsCall (liftBtns);
  }

  MvSet computeFloorIsLandingCall ()
  {
    return computeFloorIsCall (main.getLandingBtns ());
  }
  
  MvSet computeFloorIsCall (ButtonR[] btns)
  {
    MvSet dirIsDown = cur (DIR, bot);
    MvSet dirIsUp = cur (DIR, top);
    
    
    CaseStatement case1 = new CaseStatement ();
    for (int i = topFloor; i > botFloor; i--)
      case1.addCase (btns [i - 1].getPressed ().and
		     (computeFloorGreater (i - 1)),
		     computeFloorEquals (i));
    
    case1.addCase (btns [0].getPressed (), 
		   computeFloorEquals (botFloor));
    case1.addDefault (ddManager.createConstant (bot));

    
    CaseStatement case2 = new CaseStatement ();
    for (int i = botFloor + 1; i <= topFloor; i++)
      case2.addCase (btns [i - 2].getPressed ().and
		     (computeFloorLess (i)),
		     computeFloorEquals (i - 1));
    case2.addCase (btns [btns.length - 1].getPressed (),
		   computeFloorEquals (topFloor));
    case2.addDefault (ddManager.createConstant (bot));

    return dirIsDown.and (case1.compute ()).or 
      (dirIsUp.and (case2.compute ()));

  }
  


  MvSet computeFloorIsFloor ()
  {
    MvSet result = ddManager.createConstant (bot);
    
    for (int i = botFloor; i <= topFloor; i++)
      result = result.or 
	(computeNatEquals (floorBaseIdx, floorWidth, i).and
	 (computeNextNatEquals (floorBaseIdx, floorWidth, i)));
    return result;
  }
  MvSet computeFloorIsFloorPlusOne ()
  {
    MvSet result = ddManager.createConstant (bot);
    
    for (int i = botFloor; i < topFloor; i++)
      result = result.or
	(computeNatEquals (floorBaseIdx, floorWidth, i).and
	 (computeNextNatEquals (floorBaseIdx, floorWidth, i+1)));
    return result;
  }
  MvSet computeFloorIsFloorMinusOne ()
  {
    MvSet result = ddManager.createConstant (bot);
    for (int i = topFloor; i > botFloor; i--)
      result = result.or
	(computeNatEquals (floorBaseIdx, floorWidth, i).and
	 (computeNextNatEquals (floorBaseIdx, floorWidth, i-1)));
    return result;
  }
  

  

  MvSet computeCarCallEquals (int num)
  {
    return computeCallEquals (0, liftBtns);
  }
  
  MvSet computeLandingCallEquals (int num)
  {
    return computeCallEquals (0, main.getLandingBtns ());
  }
  
  MvSet computeCallEquals (int num, ButtonR[] btns)
  {
    if (num != 0) Assert.assert ("Can only compute 'car_call = 0'");
    
    MvSet dirIsDown = cur (DIR, bot);
    MvSet dirIsUp = cur (DIR, top);
    
    MvSet cond1;
    MvSet cond2;

    cond1 = btns [0].getPressed ();
    
    // -- compute a disjunction of 
    // -- butX.PRESSED & floor > (X-1)
    for (int i = botFloor; i < topFloor; i++)
      cond1 = cond1.or 
	(computeFloorGreater (i).and
	 (btns [i].getPressed ()));
    
    
    cond2 = btns [btns.length - 1].getPressed ();

    for (int i = botFloor + 1; i <= topFloor; i++)
      cond2 = cond2.or
	(computeFloorLess (i).and
	 (btns [i-2].getPressed ()));
    

    return dirIsDown.and (cond1.not ()).or (dirIsUp.and (cond2.not ()));
  }
  
  
  /***
   * DEFINE
   * idle := no_call & \A{i} !liftBut_i.pressed
  */
  MvSet computeIdle ()
  {
    MvSet result = noCall;
    
    for (int i = 0; i < liftBtns.length; i++)
      result = result.and (liftBtns [i].getPressed ().not ());
    return result;    
  }

  public MvState initializeInitState (MvState _initState)
  {
    MvState initState = super.initializeInitState (_initState);
    
    initState.setValue (getVarOrder (DOOR), bot);
    initState.setValue (getVarOrder (DIR), bot);
    initState.setValue (getVarOrder (TTFULL), bot);
    
    // -- set 'init (floor) = 1'
    initState.setValue (getVarOrder (floorBaseIdx), top);
    for (int i = 1; i < floorWidth; i++)
      initState.setValue (getVarOrder (floorBaseIdx + i), bot);

    return initState;
  }

}




