package edu.toronto.cs.modelchecker;

import edu.toronto.cs.util.*;
import edu.toronto.cs.util.userprompt.*;
import edu.toronto.cs.util.dgraph.*;
import edu.toronto.cs.prettygraph.*;
//import edu.toronto.cs.xkripke.*;
//import edu.toronto.cs.xkripke.XKripke.*;
//import edu.toronto.cs.algebra.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.mdd.*;
import edu.toronto.cs.ctl.*;
import edu.toronto.cs.smv.*;
import edu.toronto.cs.keg.*;

import edu.toronto.cs.tlq.UpSetLattice;

import java.util.*;
import java.io.*;

/**
 ** This is the front end of the model-checker and the KEG. It deals
 ** with the user supplied parameters and starts the modelchecker
 ** engine or the KEG.
 **/ 
public class XCheckFrontEnd
{

  // options
  Map options;

  // tells whether keg mode is on or not
  boolean kegmode;

  // initialize the logger
  static Logger log = new Logger (XCheckFrontEnd.class.getName ());

  // model-checker initializer
  MCManager mcconf;

  // KEG initializer
  KEGManager kegconf;
  
  // verbosity level: 0 - lowest, 2 - highest (deafult)
  int verbose;

  // initial state
  MvState initState;

  // the macro expander for handling state names
  //  StringMacroManager macro;
  CTLMacroManager macroManager;

  //  PlaceholderManager2 placeHolderManager = null;

  // -- enumerated types that we know about
  EnumeratedTypes enumTypes = EnumeratedTypes.EMPTY_TYPES;

  // list of fairness conditions
  //  List fairness;

  /**
   ** Constructor for the XCheck front end.
   **/
  public XCheckFrontEnd (Map opts)
  {
    // get the options
    options = opts;
    log.info ("Options passed to the modelchecker: " + opts);

    kegmode = ((Boolean) options.get ("kegmode")).booleanValue ();

    // initialize the modelchecker
    mcconf = new MCManager (opts);
    // initialize KEG if needed
    if (kegmode) kegconf = new KEGManager (opts, mcconf);

    verbose = mcconf.getVerbosity ();

    // get the enumerated types
    enumTypes = mcconf.getEnumTypes ();

    // -- create a macro manager
    macroManager = mcconf.getCTLMacroManager ();
    //    fairness = mcconf.getFairness ();
  }

  /**
   ** The driver for the xcheck.
   **/
  public static void main (String[] args)
  {
    Map opts = XClapi.parseOpts (args);
    System.out.println ("Options: " + opts);

    XCheckFrontEnd frontend = new XCheckFrontEnd (opts);
    frontend.run ();
  }

  /**
   ** Drives the KEG/model checker.
   **/
  private void run ()
  {
    NiceReader input = new NiceReader (System.in);    
    // -- tell the converter what lattice it should use
    BinaryTreeToCTLConverter.AssignLattice (mcconf.getMC ().getLattice ());
    
    boolean run = true;
    String work = "";

    if (kegmode) { work = "KEG"; }
    else { work = "model check"; }

    do
      {
        try
	  {
	    run = processOneProp (input);
	  }
	catch (Exception e)
	  {  
	    System.out.println ("Error while " + work + "ing: " + e);
	    e.printStackTrace ();
	    // -- don't quit if we have a problem, just move on
	    run = true;
	  }
      }
    while (run);
  }
  

  /**
   ** Reads a single property from standard input, runs the model
   ** checker or KEG, and outputs the result. This functions is called
   ** continuously from the main loop.
   **/
  private boolean processOneProp (NiceReader input)
  {
    String ctlStr = processCommand (input);

    if (ctlStr.equals ("*DONE")) // some command was executed
      {	
        return true;
      }
    else if (ctlStr.equals ("*END")) // exit the loop
      {
        return false;
      }
    else
      {
        CTLNode ctl;

        if (verbose > 0) // XXX -- V: HACK! change to constants soon . . .
          System.out.println ("Parsing: " + ctlStr);

        // -- convert the ctlStr into a CTL tree
        ctl = BinaryTreeToCTLConverter.convertToCTL (ctlStr,
						     mcconf.getFairness ());
        // -- expand the macroes -- called twice for 2 pass expansion
        ctl = macroManager.expandMacro (ctl);
        ctl = macroManager.expandMacro (ctl);

        if (verbose > 0) // XXX -- V: HACK! change to constants soon . . .
          System.out.println ("After macro expansion: " + ctl);
	
	if (kegmode) { kegCTL (ctl); }
	else { checkCTL (ctl); }

	return true;
      }    
  }

  private void kegCTL (CTLNode ctl)
  {
    DGraph KGraph = kegconf.doCtl (ctl);

    // get the filename from the user
    NiceReader reader = new NiceReader (System.in);
    
    System.out.println
      ("Please enter filename; hit enter for automatic. ("
       + kegconf.getAutoName () + ")");
    
    String filename = reader.readLine ();
    
    if (filename.equals (""))
      filename = kegconf.getAutoName ();
    // use default filename

    // Choose colourmap
    Map colourmap = edu.toronto.cs.prettygraph.PrettyConstants.singletonMap;

    kegconf.saveKegToFile (KGraph, filename, colourmap);
  }

  private void checkCTL (CTLNode ctl)
  {
    PropOrderIf propOrder = mcconf.getPropOrder ();

    // -- run our model checker on the property
    // -- record the time that it takes
    StopWatch stopwatch = new StopWatch ();
    MvSet mvResult = mcconf.getMC ().doCTL (ctl);

    stopwatch.stop ();
	
    if (verbose > 0) // this will do for now
      {
	System.out.println ("time: " +
			    stopwatch.getTimeElapsed ()/1000.0 + "s");
	System.out.println ("size of trans: " +
			    mcconf.getMC ().getMvTransRelation ().size ());
	System.out.println (mvResult.getFactory ().toString ());    
      }

    // -- print the result
    if (verbose > 0) // XXX -- V: HACK! change to constants soon . . .
      System.out.println
	("This is the formula that you are checking: " +
	 macroManager.compressMacro (ctl));

    //          if (mvResult instanceof MDDMvSet)
    //            writeMDDToFile ((MDDMvSet) mvResult);

    //          System.out.println ("Final result would've been here..." +
    //                              " disabled for now");
    // XXX
    printResult (mvResult);
    System.out.println ("Writing result to a file");
    mvResult.writeToDaVinci ("/tmp/result.daVinci", propOrder);
  }
  

  /**
   ** Reads a single command (either a CTL formula or a command to
   ** define a macro or switch state).
   ** @return -- a String representing a CTL formula, "*DONE", or "*END".
   **/
  private String processCommand (NiceReader input) 
  {
    // stuff needed to read the CTL formulas
    String ctlStr;
    // prompt
    String prompt = "";
    
    if (kegmode) 
      { prompt = "KEG:> "; }
    else
      { prompt = "MC:> "; }
    

    // -- reset macro manager each time
    macroManager.reset ();

    // print the propositional ordering so that it is easier to
    // construct a CTL
    printPropOrdering ();

    // read a CTL
    System.out.println
      (prompt + "What CTL property would you like to check today?");
    ctlStr = input.readLine ();

    if (ctlStr.length () == 0) return "*END"; // empty input - exit the loop
    else if (ctlStr.equals ("state"))
      {
        System.out.println ("Type in the name of the new initial state" +
                            "(leave blank to keep the old one):");
        String name = input.readLine ();
        options.put ("initstate", name);
	mcconf.passOptions (options);
        initState = mcconf.constructInitialState ();
      }
    else if (ctlStr.equals ("macro"))
      {
        System.out.println ("Type in the name of the new macro:");
        String name = input.readLine ();
        System.out.println ("Type in the expansion of the new macro:");
        String expansion = input.readLine ();
        // -- transform into the standard CTL String representation
        CTLNode exp = BinaryTreeToCTLConverter.convertToCTL 
	  (expansion, mcconf.getFairness ());

        macroManager.defineMacro (CTLFactory.createCTLAtomNode (name),
                                  exp);
      }
    else if (ctlStr.equals ("macro_reset"))
      {
        System.out.println ("Reseting macros");
        macroManager.reset ();
      }
    else if (ctlStr.equals ("fairness"))
      {
	System.out.println ("Fairness conditions are: " + printFairness ());
        System.out.println ("Type in the fairness condition(s)" +
			    " separated by commas:");
        String condition = input.readLine ();
	
	for (StringTokenizer strtok = new StringTokenizer (condition, ",");
	     strtok.hasMoreTokens ();)
	  {
	    String cond = strtok.nextToken ();

	    if (cond.length () > 0)
	      {
		// -- transform into the standard CTL String representation
		CTLNode c = BinaryTreeToCTLConverter.
		  convertToCTL (cond, new LinkedList());

		// -- expand the macroes -- called twice for 2 pass expansion
		c = macroManager.expandMacro (c);
		c = macroManager.expandMacro (c);

		mcconf.getFairness ().add (c);
	      }
	  }
	
      }
    else if (ctlStr.equals ("fairness_reset"))
      {
        System.out.println ("Reseting fairness conditions");
	mcconf.setFairness (new LinkedList ());
      }
    // everything else is treated as a CTL formula
    else
      {
	return ctlStr;
      }
    return "*DONE";
  }


  /**
   ** Prints the result of the model-checking restricted to the
   ** initial state.
   **/
  private void printResult (MvSet mvResult)
  {
    String res = mcconf.resultToString (mvResult);
    // depending on the verbosity print additional info
    PrintWriter out = new PrintWriter (System.out);
    switch (verbose)
      {
      case 2:
        // -- print a map that maps user's idea about logic values and
        // -- internal representation used in debug print outs
        mcconf.getMC ().getLattice ().dumpValueMap (out);
      case 1: // nothing for now
      case 0: break;
      }
    out.flush ();
    System.out.println (res);
  }


  /**
   ** Prints the current propositional ordering.
   **/
  private void printPropOrdering ()
  {
    PropOrderIf propOrder = mcconf.getPropOrder ();

    if (verbose > 1)
      {
	int size = propOrder.size ();
	for (int i = 0; i < size; i++)
	  {
	    if (propOrder.isPostImage (i))
	      {
		if (verbose > 1)
		  System.out.println
		    ("#" + i + " = " + propOrder.getPropName (i) + "'");
	      }
	    else
	      System.out.println
		("#" + i + " = " + propOrder.getPropName (i));
	  }
      }
  }

  /**
   ** Makes sure that the fairness conditions are printed in a
   ** human-readable format.
   **/
  private String printFairness ()
  {
    String output;
    
    output = "{";
    for (Iterator i = mcconf.getFairness ().iterator (); i.hasNext ();)
      {
	CTLNode ctl = (CTLNode) i.next ();

	output = output + macroManager.compressMacro (ctl);
	if (i.hasNext ()) output = output + ", ";
      }
    output = output + "}";
    return output;
  }
}
