package edu.toronto.cs.modelchecker;

import edu.toronto.cs.util.*;
import edu.toronto.cs.xkripke.*;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.smv.*;

import java.util.*;
import java.lang.reflect.Method;


/**
 ** A reader for XML models.
 **/
public class XMLModelReader extends AbstractModelReader
{
    // initialize the logger
  static Logger log = new Logger (XMLModelReader.class.getName ());

  public XMLModelReader (Map opts)
  {
    options = opts;
    // set up the default model name
    options.put ("model name", ((List) options.get ("input")).get (0));

    // generate the xkripke structure from the input file(s)
    makeXKripke ();
	
    // -- the following components are needed to initialize 
    // -- the modelchecker
    // create the propositional ordering
    makePropOrder ();
  }  


  /**
   ** Given an MvSetFactory creates a transition relation.
   **/
  public void makeTransRelation (MvSetFactory ddManager)
  {
    // get the MvSet representation of the transition relation
    transRelation = createTransRelation (ddManager);
	
    // get the initial state
    initState = constructInitialState ();
    //System.out.println ("The initial state is: " + initState);
  }

  /**
   ** Generates the xkripke structure(s) from the input file(s) specified
   ** in the options.
   **/
  private void makeXKripke ()
    throws IllegalArgumentException, NestedRuntimeException
  {
    // get the file name(s)
    List filenames = (List) options.get ("input");

    // determine whether states should be labelled
    boolean addStateNames = ((Boolean) options.get ("statenames")).
      booleanValue ();

    if (filenames.size () == 0)
      {
	throw new IllegalArgumentException 
	  ("Cannot create XKripke structure: no input file specified.");
      }
    else
      {
	try
	  {
	    xkripke = XKripkeFactory.parse 
	      ((String []) filenames.toArray ());
	  }
	catch (Exception e)
	  {
	    throw new NestedRuntimeException (e);
	  }
      }

    // add state name labels if needed
    if (addStateNames)
      for (int x = 0; x < xkripke.length; x++) xkripke [x].labeliseStates ();
  }

  /**
   ** Creates the appropriate propositional ordering for all the
   ** XKripke structires.
   **/
  private void makePropOrder ()
  {
    // first get all the propositional variables in one set
    List allprops = new ArrayList ();
    
    for (int x = 0; x < xkripke.length; x++)
      allprops.addAll (xkripke [x].getPropNamesAsSet ());
    // now feed them to the AlternatePropOrder constructor
    propOrder = new AlternatePropOrder ((String []) allprops.toArray 
					(new String [allprops.size ()]));
  }
  /**
   ** Constructs the transition relation represented as an MvSet from
   ** the xkripke structure.
   **
   ** @return -- the transition relation corresponding to the
   ** specified xkripke structure
   **/

  private MvSet createTransRelation (MvSetFactory mvSetFactory)
  {
    // construct the MvSet using the appropriate factory
    return XKripkeMvSetBuilder.buildMvSet (mvSetFactory, xkripke,
					    propOrder);
  }
  /**
   ** Gets the initial state for the current model.
   **
   ** @return -- the initial state for the model or null if no unique
   ** initial state was defined.
   **/
  public MvState constructInitialState ()
  {
    // get the name of the initial state from the options
    String name = (String) options.get ("initstate");
    FiniteLattice lattice = xkripke [0].getLattice ();
    MvState initstate = new MvState (propOrder.size (), lattice);
     MvState [] states = new MvState [xkripke.length];

    if (name == null) // no initial state was specified - use default
      {
	// unify all the initial states
	for (int x = 0; x < xkripke.length; x++)
	  {
	    // get the initial state of the current model
	    states [x] = new MvState (xkripke [x].getInitialStates () [0],
				      propOrder);
	  }
      }
    else // a particular state was selected
      {
	StringTokenizer namelist = new StringTokenizer (name, ",");

	for (int x = 0; x < xkripke.length; x++)
	  states [x] = new MvState (xkripke [x].getState
				      (namelist.nextToken ()), propOrder);
      }

    // copy all the valid values of the propositions
    for (int x = 0; x < xkripke.length; x++)
      for (int i = 0; i < initstate.size (); i++)
	if (states [x].getValue (i) != Algebra.NO_VALUE)
	  initstate.setValue (i, states [x].getValue (i));

    return initstate;
  }

  /**
   ** Get the lattice.
   ** @return lattice.
   **/
  public FiniteLattice getLattice ()
  {
    return xkripke[0].getLattice ();
  }
 
}

