package edu.toronto.cs.mvset;

import edu.toronto.cs.mdd.*;
import edu.toronto.cs.util.*;
import edu.toronto.cs.algebra.*;
import java.util.Arrays;
import edu.toronto.cs.modelchecker.PropOrderIf;

import java.io.FileWriter;

/***
 *** Implementation of an MvSet based on the MDD library
 *** and join irreducible decomposition
 ***/
public class JMvSet extends AbstractMvSet
{

  JMvSetFactory factory;
  
  // -- manager for the underlying MDDs
  MDDManager mddManager;

  // -- Finite lattice corresponding to the manager
  FiniteLattice lattice;
  
  // -- the actuall MDDs representing this function
  MDDNode[] mdd;
  // -- join irreducibles of the lattice such that
  // -- mdd [i] represents an l-cut w.r.t. joinIrr [i]
  int[] joinIrr;
  int[] invJoinIrr;
  
  // -- negation table
  int[] neg;
   
  public JMvSet (JMvSetFactory _factory, MDDNode[] _mdd)
  {
    mdd = _mdd;
    factory = _factory;
    mddManager = factory.getMddManager ();
    joinIrr = factory.getJoinIrr ();
    invJoinIrr = factory.getInvJoinIrr ();
    neg = factory.getNegTable ();
    lattice = factory.getLattice ();
  }
  




  /*** 
   *** Creates a pointwise composition
   *** op is an operator: L x L -> L
   *** f is the current mv-set
   *** result h (x) = f (x) op g (x)
   ***/
  public MvSet ptwiseCompose (int op, MvSet g)
  {
    MDDNode[] result = new MDDNode [joinIrr.length];

    MDDNode[] gSets = ((JMvSet)g).getMddNodes ();

    for (int i = 0; i < result.length; i++)
      result [i] = mddManager.Apply (mdd [i], gSets [i], op);
    
    
    return createMvSet (result);
  }
  
  public MvSet ptwiseCompare (int op, MvSet g)
  {
    if (op != FiniteLattice.ABOVE &&
	op != FiniteLattice.BELOW &&
	op != FiniteLattice.EQUAL)
      Assert.assert ("Cannot comapre using: " + op);

    MDDNode node = mddManager.getOne ();
    
    MDDNode[] gSets = ((JMvSet)g).getMddNodes ();

    for (int i = 0; i < gSets.length; i++)
      node = mddManager.Apply (node, 
			       mddManager.Apply (mdd [i], 
						 gSets [i], 
						 op), FiniteLattice.MEET);
    
    MDDNode[] result = new MDDNode [joinIrr.length];
    Arrays.fill (result, node);
    return createMvSet (result);
  }  

 
  /***
   *** Creates a ptwise negation
   *** h (x) = \neg f (x)
   ***/
  public MvSet ptwiseNeg ()
  {
    MDDNode[] mddNodes = new MDDNode [joinIrr.length];
    
    // -- mddNodes [i] is the result of applying a lattice NEG
    // -- to the mddNodes [j] where j corresponds to the neg table value
    // -- of join irreducible corresponding to mdd [i]
    for (int i = 0; i < joinIrr.length; i++)
      mddNodes [i] = mddManager.Apply (mdd [invJoinIrr [neg [joinIrr [i]]]],
					 FiniteLattice.NEG);	
    return createMvSet (mddNodes);
  }
  
  public int size ()
  {
    return mddManager.sharedSize (mdd);
  }

  /***
   *** Restricts an argument
   *** h (x_0, x_1, x_2, ...) = f (x_0, ..., value, ...)
   *** where value is substituted at argIdx
   ***/
  public MvSet restrict (int argIdx, int value)
  {
    MDDNode[] result = new MDDNode [joinIrr.length];
    for (int i = 0; i < result.length; i++)
      result [i] = mddManager.Restrict (mdd [i], argIdx, value);
    return createMvSet (result);
  }
  
  public MvSet restrict (MvState state)
  {
    MDDNode[] result = new MDDNode [joinIrr.length];

    System.arraycopy (mdd, 0, result, 0, mdd.length);
    
    for (int i = 0; i < result.length; i++)
      for (int j = state.getFirstVar (); j <= state.getLastVar (); j++)
	if (state.isSet (j))
	  result [i] = mddManager.Restrict (result [i], j, 
					    state.getValue (j));
    return createMvSet (result);
  }
  

  /***
   *** Renames the arguments. newArgs is a map from old args to new so that
   *** h (x) = f (newArgs [0], newArgs [1], ...)
   ***/
  public MvSet renameArgs (int[] newArgs)
  {

    MDDNode[] result = new MDDNode [joinIrr.length];
    for (int i = 0; i < result.length; i++)
      result [i] = mddManager.renameVars (mdd [i], newArgs);

    return createMvSet (result);
  }

  public MvSet existAbstract (MvSet _cube)
  {
    MDDNode cube = ((JMvSet)_cube).getMddNodes () [0];
    
    return createMvSet (existAbstract (cube));
  }

  MDDNode[] existAbstract (MDDNode cube)
  {
    MDDNode[] mdds = new MDDNode [joinIrr.length];
    for (int i = 0; i < mdds.length; i++)
      mdds [i] = mddManager.existAbstract (mdd [i], cube);
    return mdds;
  }
  
  public MvSet existAbstract (int[] varIndex)
  {
    MDDNode cube = buildCube (varIndex);
    return createMvSet (existAbstract (cube));
  }

  MDDNode buildCube (int[] varIndex)
  {
    int top = lattice.idtimesIdx ();
    MDDNode cube = mddManager.getOne ();
    for (int i = 0; i < varIndex.length; i++)
      cube = mddManager.Apply (cube, 
			       mddManager.getVarEquals (varIndex [i], top),
			       BooleanLattice.MEET);
    return cube;
  }
  
  
  public MvSet forallAbstract (MvSet _cube)
  {
    MDDNode cube = ((JMvSet)_cube).getMddNodes () [0];
    
    return createMvSet (forallAbstract (cube));
  }
  
  MDDNode[] forallAbstract (MDDNode cube)
  {
    MDDNode[] result = new MDDNode [joinIrr.length];
    for (int i = 0; i < result.length; i++)
      result [i] = mddManager.forallAbstract (mdd [i], cube);
    return result;
  }
  
  public MvSet forallAbstract (int[] varIndex)
  {
    MDDNode cube = buildCube (varIndex);
    return createMvSet (forallAbstract (cube));
  }
  
  
  /***
   *** Evaluates this function on an input
   *** result = f (values [0], values [1], ...)
   ***/
  public int evaluate (int[] values)
  {
    // XXX Note: we will not use MvSet restrict operation here
    // XXX Note: to avoid creating a lot of intermidiate MvSets
    MDDNode[] mddNodes = new MDDNode [joinIrr.length];
    
    // -- copy the current array of MDDNodes
    System.arraycopy (mdd, 0, mddNodes, 0, mdd.length);
    
    // -- first restrict each mdd individually
    for (int i = 0; i < values.length; i++)
      for (int j = 0; j < mddNodes.length; j++)
	mddNodes [j] = mddManager.Restrict (mddNodes [j], i, values [i]);
    
    // -- a final step in evaluation is to compute a join of all 
    // -- join irreducibles whose MDDNode is One (or Top)

    
    // -- start with a lowest possible answer
    int value = lattice.idplusIdx ();
    
    for (int i = 0; i < mddNodes.length; i++)
      if (mddNodes [i] == mddManager.getOne ())
	value = lattice.plus (value, joinIrr [i]);
    return value;
  }

  public int evaluate (MvState state)
  {
    MvSet restricted = this.restrict (state);
    MDDNode[] mddNodes = ((JMvSet)restricted).getMddNodes ();
    
    int value = lattice.idplusIdx ();
    
    for (int i = 0; i < mddNodes.length; i++)
      if (mddNodes [i] == mddManager.getOne ())
	value = lattice.plus (value, joinIrr [i]);
    return value;
  }
  


  /*** Arrity of this function */
  public int getArrity ()
  {
    return factory.getArrity ();
  }

  /*** Gets the lattice */
  public FiniteLattice getLattice ()
  {
    return lattice;
  }

  public MDDNode[] getMddNodes ()
  {
    return mdd;
  }

  protected MvSetFactory getMvSetFactory ()
  {
    return factory;
  }
  
  private MvSet createMvSet (MDDNode[] mddNode)
  {
    return factory.createMvSet (mddNode);
  }
  

  public boolean equals (Object o)
  {
    if (! (o instanceof JMvSet)) return false;
    return equals ((JMvSet)o);
  }
  
  public boolean equals (JMvSet v)
  {
    // -- two mv-sets can only be equal if they come from the same
    // -- factory
    if (factory != v.getMvSetFactory ())
      return false;

    // -- two mv-sets are equal if their mdd representations are equal
    MDDNode[] gSets = v.getMddNodes ();
    for (int i = 0; i < mdd.length; i++)
      if (! (mdd [i].equals (gSets [i]))) return false;
    return true;    
  }
  

  public static MvSetFactory newMvSetFactory (FiniteLattice lattice, 
					      int arrity)
  {
    return new JMvSetFactory (lattice, arrity);
  }

  public void writeToDaVinci (String fileNamePrefix, PropOrderIf propOrder)
  {
    try {
      for (int i = 0; i < mdd.length; i++)
	{
	  String fileName = fileNamePrefix + 
	    lattice.getElementKeyByIdx (joinIrr [i]) + ".daVinci";
	  
	  FileWriter out = new FileWriter (fileName);
	  
	  out.write (mdd [i].toDaVinciString (propOrder));
	  out.flush ();
	  out.close ();
	}
    } catch (Exception ex) {
      Assert.assert (ex);
    }
  }


  

  public MvSetFactory getFactory ()
  {
    return factory;
  }  

  public static class JMvSetFactory extends AbstractMvSetFactory
  {
    int arrity;
    MDDManager mddManager;
    FiniteLattice lattice;

    int[] joinIrr;
    // -- this is the inverse of joinIrr
    // -- let j be a join irreducible 
    // -- invJoinIrr [j] == i -> joinIrr [i] == j
    // -- it's size is the size of the lattice, 
    // -- and it is not defined for non-join irreducible elements
    int[] invJoinIrr;
    int[] neg;
    
    public JMvSetFactory (FiniteLattice _lattice,
			  int _arrity)
    {
      arrity = _arrity;
      lattice = _lattice;
      mddManager = new MDDMgr (arrity, lattice);
      joinIrr = lattice.getJoinIrreducibles ();

      invJoinIrr = computeInvJoinIrr (lattice.size (), joinIrr);
      
      // -- create the negation table
      neg = new int [lattice.size ()];
      computeNegTable ();
      
    }
    
    private int[] computeInvJoinIrr (int size, int[] joinIrr)
    {
      int[] invJoinIrr = new int [size];
      for (int i = 0; i < joinIrr.length; i++)
	invJoinIrr [joinIrr [i]] = i;
      return invJoinIrr;
    }
    

    /***
     *** Given a value in L creates a function 
     *** f(x_0, x_1, ...) = value
     ***/
    public MvSet createConstant (int value)
    {
      MDDNode[] mddNodes = new MDDNode [joinIrr.length];

      for (int i = 0; i < mddNodes.length; i++)
	{
	  if (lattice.leq (joinIrr [i], value))
	    mddNodes [i] = mddManager.getOne ();
	  else
	    mddNodes [i] = mddManager.getZero ();
	}
      

//        Arrays.fill (mddNodes, mddManager.getZero ());
      
//        int[] decomp = lattice.joinDecomposition (value);
//        for (int i = 0; i < decomp.length; i++)
//  	mddNodes [invJoinIrr [decomp [i]]] = mddManager.getOne ();

      return createMvSet (mddNodes);
    }    
    
    /***
     *** Given an argument index (0 <= argIdx < n)
     *** constructs a projection function
     *** f(x_0, x_1, ...) = x_argIdx
     ***/
    public MvSet createProjection (int argIdx)
    {
      MDDNode[] mddNodes = new MDDNode [joinIrr.length];
      /**
       ** Each mddNodes [i] is an MDD representing the following function
       ** Let j = joinIrr [i]
       ** Let A = up set of j
       ** mddNodes [i] (x) = \/ (x = a) where a ranges over A
       **/
      
      for (int i = 0; i < mddNodes.length; i++)
	mddNodes [i] = 
	  createVarEqMDD (argIdx, lattice.getUpSet (joinIrr [i]));
      
      
      return createMvSet (mddNodes);
    }

    private MDDNode createVarEqMDD (int varIdx, int[] values)
    {
      MDDNode mddNode = mddManager.getZero ();
      for (int i = 0; i < values.length; i++)
	mddNode = 
	  mddManager.Apply (mddNode, 
			    mddManager.getVarEquals (varIdx, values [i]), 
			    FiniteLattice.JOIN);
      return mddNode;
    }
    

    
    /***
     *** Creates a point function 
     *** f (args) = value
     ***          = 0 otherwise
     ***/
    public MvSet createPoint (int[] args, int value)
    {
      MDDNode[] mddNodes = new MDDNode [joinIrr.length];
      

      

      MDDNode mddNode = mddManager.getOne ();
      for (int i = 0; i < args.length; i++)
	{
	  if (args [i] == MvSetFactory.DONT_CARE ||
	      args [i] == Algebra.NO_VALUE) continue;
	  mddNode = mddManager.Apply (mddNode, 
				      mddManager.getVarEquals (i, args [i]),
				      FiniteLattice.MEET);
	}
      
      // -- we will now place the resulting mddNode in the appropriate
      // -- places
      for (int i = 0; i < mddNodes.length; i++)
	  {
	      if (lattice.leq (joinIrr [i], value))
		  mddNodes [i] = mddNode;
	      else
		  mddNodes [i] = mddManager.getZero ();
	  }
      
      //Arrays.fill (mddNodes, mddManager.getZero ());      
//        int[] decomp = lattice.joinDecomposition (value);
//        for (int i = 0; i < decomp.length; i++)
//  	mddNodes [invJoinIrr [decomp [i]]] = mddNode;
      return createMvSet (mddNodes);
    }

    public MvSet createPoint (MvState state, int value)
    {
      MDDNode[] mddNodes = new MDDNode [joinIrr.length];
      

      

      MDDNode mddNode = mddManager.getOne ();
      for (int i = state.getFirstVar (); i <= state.getLastVar (); i++)
	if (state.isSet (i))
	  mddNode = mddManager.Apply 
	    (mddNode, 
	     mddManager.getVarEquals (i, state.getValue (i)),
	     FiniteLattice.MEET);

      // -- we will now place the resulting mddNode in the appropriate
      // -- places
      for (int i = 0; i < mddNodes.length; i++)
	{
	  if (lattice.leq (joinIrr [i], value))
	    mddNodes [i] = mddNode;
	  else
	    mddNodes [i] = mddManager.getZero ();
	}

//        Arrays.fill (mddNodes, mddManager.getZero ());
//        int[] decomp = lattice.joinDecomposition (value);
//        for (int i = 0; i < decomp.length; i++)
//  	mddNodes [invJoinIrr [decomp [i]]] = mddNode;
      return createMvSet (mddNodes);
    }


    public MvSet var (int argIdx, int argVal, int value)
    {
      MDDNode mddNode = mddManager.getOne ();
      mddNode = mddManager.Apply (mddNode, 
				  mddManager.getVarEquals (argIdx, argVal),
				  FiniteLattice.MEET);
      
      MDDNode[] mddNodes = new MDDNode [joinIrr.length];
      
      for (int i = 0; i < mddNodes.length; i++)
	{
	  if (getLattice ().leq (joinIrr [i], value))
	    mddNodes [i] = mddNode;
	  else
	    mddNodes [i] = mddManager.getZero ();
	}
      return createMvSet (mddNodes);
    }
    

    
    public MvSet createMvSet (MDDNode[] mddNodes)
    {
      return new JMvSet (this, mddNodes);
    }

    public FiniteLattice getLattice ()
    {
      return lattice;
    }
    
    public MDDManager getMddManager ()
    {
      return mddManager;
    }
    
    protected int getArrity ()
    {
      return arrity;
    }
    
    protected int[] getJoinIrr ()
    {
      return joinIrr;
    }
    protected int[] getInvJoinIrr ()
    {
      return invJoinIrr;
    }
    
    protected int[] getNegTable ()
    {
      return neg;
    }

    private void computeNegTable ()
    {
      int[] f = computeF ();
      
      for (int i = 0; i < joinIrr.length; i++)
	{
	  int j = joinIrr [i];
	  neg [lattice.unOp (f [j], FiniteLattice.NEG)] = j;
	}      
    }
    private int[] computeF ()
    {
      int[] f = new int [lattice.size ()];
      
      for (int i = 0; i < joinIrr.length; i++)
	f [joinIrr [i]] = computeF (joinIrr [i]);
      return f;
    }
    private int computeF (int j)
    {
      // -- return max (L - upSet (j))
      int[] upSet = lattice.getUpSet (j);
      
      int result = lattice.idplusIdx ();
      
      for (IntIterator it = lattice.carrierSet (); it.hasNext ();)
	{
	  int x = it.nextInt ();
	  if (lattice.eq (x, j)) continue;
	  // x < j ==> x \in (L - up (j))
	  else if (lattice.leq (x, j)) 
	    result = lattice.plus (result, x);
	  // x incomparable to j ==> x \in (L - up (j))
	  else if (!lattice.geq (x, j)) 
	    result = lattice.plus (result, x);
	}
      
      return result;
    }
    
    
    
  }
  

}





  
  
