package edu.toronto.cs.mvset;

import edu.toronto.cs.mdd.*;
import edu.toronto.cs.algebra.*;

import java.io.*;
import java.util.*;

import edu.toronto.cs.util.dgraph.*;
import edu.toronto.cs.prettygraph.*;
import edu.toronto.cs.modelchecker.PropOrderIf;
import edu.toronto.cs.util.Assert;
import edu.toronto.cs.util.StopWatch;
import edu.toronto.cs.util.IntIterator;


/***
 *** Implementation of an MvSet based on the MDD library
 ***/
public class MDDMvSet extends AbstractMvSet
{

  MDDMvSetFactory factory;
  
  // -- manager for the underlying MDDs
  MDDManager mddManager;
  
  // -- the actuall MDD representing this function
  MDDNode mdd;

  // -- the lattice associaterd with the MvSet
  FiniteLattice lattice;
  
  public MDDMvSet (MDDMvSetFactory _factory, MDDNode _mdd)
    {
      mdd = _mdd;
      factory = _factory;
      mddManager = factory.getMddManager ();
      lattice = factory.getLattice ();
    }

  /*** 
   *** Creates a pointwise composition
   *** op is an operator: L x L -> L
   *** f is the current mv-set
   *** result h (x) = f (x) op g (x)
   ***/
  public MvSet ptwiseCompose (int op, MvSet g)
  {
    if (FiniteLattice.JOIN == op)
      factory.countAnd ();
    else if (FiniteLattice.MEET == op)
      factory.countOr ();
    else
      factory.countOther ();
    
    MDDNode node = mddManager.Apply (mdd, ((MDDMvSet)g).getMddNode (), op);
    
//      return createMvSet (mddManager.Apply 
//  			(mdd, ((MDDMvSet)g).getMddNode (), op));
    if (FiniteLattice.JOIN == op)
      factory.endCountAnd ();
    else if (FiniteLattice.MEET == op)
      factory.endCountOr ();

    return createMvSet (node);
  }
  
  public MvSet ptwiseCompare (int op, MvSet g)
  {
    return ptwiseCompose (op, g);
  }


  /***
   *** Creates a ptwise negation
   *** h (x) = \neg f (x)
   ***/
  public MvSet ptwiseNeg ()
  {
    factory.countNeg ();
    MDDNode node = mddManager.Apply (mdd, FiniteLattice.NEG);
    //return createMvSet (mddManager.Apply (mdd, FiniteLattice.NEG));
    factory.endCountNeg ();
    return createMvSet (node);
  }
  
  public int size ()
  {
    return mddManager.dagSize (mdd);
  }


  /***
   *** Restricts an argument
   *** h (x_0, x_1, x_2, ...) = f (x_0, ..., value, ...)
   *** where value is substituted at argIdx
   ***/
  public MvSet restrict (int argIdx, int value)
  {
    return createMvSet (mddManager.Restrict (mdd, argIdx, value));
  }
  
  public MvSet restrict (MvState state)
  {
    MDDNode mddNode = mdd;
    for (int i = state.getFirstVar (); i <= state.getLastVar (); i++)
      if (state.isSet (i))
	mddNode = mddManager.Restrict (mddNode, i, state.getValue (i));
    return createMvSet (mddNode);
  }
  

  public MvSet existAbstract (MvSet _cube)
  {
    factory.countExistAbstract ();
    MDDNode cube = ((MDDMvSet)_cube).getMddNode ();
    MDDNode result = mddManager.existAbstract (mdd, cube);
    factory.endCountExistAbstract ();
    return createMvSet (result);
  }
  public MvSet forallAbstract (MvSet _cube)
  {
    factory.countForallAbstract ();
    MDDNode cube = ((MDDMvSet)_cube).getMddNode ();
    MDDNode result = mddManager.forallAbstract (mdd, cube);
    factory.endCountForallAbstract ();
    return createMvSet (result);
  }
  
  MDDNode buildCube (int[] varIndex)
  {
    int top = getLattice ().idtimesIdx ();
    
    MvState state = new MvState (getArrity (), getLattice ());
    for (int i = 0; i < varIndex.length; i++)
      state.setValue (varIndex [i], top);

    MDDNode point = 
      ((MDDMvSet)factory.createPoint (state, top)).getMddNode ();
    return point;
  }
  
  
  public MvSet existAbstract (int[] varIndex)
  {
    factory.countExistAbstract ();
    MDDNode point = buildCube (varIndex);
    //System.out.println ("existAbstract");

    MDDNode result = mddManager.existAbstract (mdd, point);
    factory.endCountExistAbstract ();
    return createMvSet (result);
  }
  
  public MvSet forallAbstract (int[] varIndex)
  {
    factory.countForallAbstract ();
    MDDNode point = buildCube (varIndex);
    MDDNode result = mddManager.forallAbstract (mdd, point);
    factory.endCountForallAbstract ();
    return createMvSet (result);
  }
  

  /***
   *** Renames the arguments. newArgs is a map from old args to new so that
   *** h (x) = f (newArgs [0], newArgs [1], ...)
   ***/
  public MvSet renameArgs (int[] newArgs)
  {
    return createMvSet (mddManager.renameVars (mdd, newArgs));
  }
  
  
  /***
   *** Evaluates this function on an input
   *** result = f (values [0], values [1], ...)
   ***/
  public int evaluate (int[] values)
  {
    MDDNode result = mdd;
    for (int i = 0; i < values.length; i++)
      result = mddManager.Restrict (result, i, values [i]);

    return result.getValue ();
  }
  

  public int evaluate (MvState state)
  {
    MvSet restricted = this.restrict (state);
    MDDNode mddNode = ((MDDMvSet)restricted).getMddNode ();
    if (mddNode.isConstant ())
      return mddNode.getValue ();
    
    return Algebra.NO_VALUE;
  }
  

  /*** arrity of this function */
  public int getArrity ()
  {
    return factory.getArrity ();
  }

  /*** get the lattice for the MvSet */
  public FiniteLattice getLattice ()
  {
    return lattice;
  }

  public MDDNode getMddNode ()
  {
    return mdd;
  }

  protected MvSetFactory getMvSetFactory ()
  {
    return factory;
  }
  
  private MvSet createMvSet (MDDNode mddNode)
  {
    return factory.createMvSet (mddNode);
  }

  public boolean equals (Object o)
  {
    if (! (o instanceof MDDMvSet)) return false;
    return equals ((MDDMvSet)o);
  }
  
  public boolean equals (MDDMvSet v)
  {
    // -- two mv-sets can only be equal if they come from the same
    // -- factory
    if (factory != v.getMvSetFactory ())
      return false;

    // -- two mv-sets are equal if their mdd representations are equal
    return getMddNode ().equals (v.getMddNode ());
  }

  public MvSetFactory getFactory ()
  {
    return factory;
  }

  public static MvSetFactory newMvSetFactory (Algebra algebra, 
					      int arrity)
  {
    return new MDDMvSetFactory (algebra, arrity);
  }
  

 //   public void writeToDaVinci (String filename, PropOrderIf propOrder)
//    {
//      try {
//        // get the MDD from the MvSet
//        MDDNode result = mdd;
//        // get the output filename
      

//        if (result != null)
//  	{
//  	  // -- write a graphical representation of resulting mdd
//  	  System.out.println ("AM SO writing mdd to " + filename);
	  
//  	  // Below should work, but does (or at least 1) creates output
//  	  // and 2) doesn't crash).
//  	  DGraph tmp = PrettyGraphMDDVisitor.getPrettyRep (result, propOrder);
	  
//  	  DVGenNodeVisitor dvg = 
//  	    new DVGenNodeVisitor (new DVGenEdgeVisitor ());

//  	  for (Iterator it = DGDepthTraverse.getIterator (tmp); 
//  	       it.hasNext ();)
//  	    dvg.visit ((DGNode)it.next (), null);
	  
	  
//  	  FileWriter dvmdd = new FileWriter (filename);
//  	  dvmdd.write (dvg.getSB ().toString ());
//  	  dvmdd.close ();
//  	}
//      } catch (Exception ex) {
//        ex.printStackTrace ();
//        Assert.assert (ex);
//      }
    

//    }
  
  public void resetStats ()
  {
    factory.resetStats ();
  }
  

  public static class MDDMvSetFactory extends AbstractMvSetFactory
  {
    int arrity;
    MDDManager mddManager;

    StopWatch existAbstractWatch = null;
    StopWatch forallAbstractWatch = null;
    StopWatch andWatch = null;
    StopWatch orWatch = null;
    StopWatch negWatch = null;
    
    int existAbstractCount = 0;
    int forallAbstractCount = 0;
    int andCount = 0;
    int orCount = 0;
    int negCount = 0;
    int otherCount = 0;
    
    public MDDMvSetFactory (Algebra algebra,
			    int _arrity)
    {  
      arrity = _arrity;
      FiniteLattice lattice = (FiniteLattice)algebra;
      
      int latSize = lattice.size () - 2;
      int[] ignoreMap;
      // XXX The code below constructs some type of ignore map for testing
      // XXX of some ideas. It should definitelly not be here
//       if (latSize == 0) ignoreMap = null;
//       else
// 	{
// 	  ignoreMap = new int [latSize];
// 	  int count = 0;
// 	  for (IntIterator it = lattice.carrierSet (); it.hasNext ();)
// 	    {
// 	      int val = it.nextInt ();
// 	      if (val == lattice.idtimesIdx () || val == lattice.idplusIdx ())
// 		continue;
// 	      else
// 		ignoreMap [count++] = val;
// 	    }
	  
// 	}
      ignoreMap = null;
      mddManager = new MDDMgr (arrity, algebra, ignoreMap);
    }
    
    public void setCaching (boolean v)
    {
      mddManager.setCaching (v);
    }
    
    void resetStats ()
    {
      existAbstractCount = 0;
      forallAbstractCount = 0;
      andCount = 0;
      orCount = 0;
      negCount = 0;
      otherCount = 0;
      andWatch = null;
      orWatch = null;
      negWatch = null;
      existAbstractWatch = null;
      forallAbstractWatch = null;
      mddManager.resetStats ();
    }
    
    void countExistAbstract ()
    {
      if (existAbstractWatch == null) existAbstractWatch = new StopWatch ();
      else existAbstractWatch.resume ();
      
      existAbstractCount++;
    }
    void endCountExistAbstract ()
    {
      existAbstractWatch.pause ();
    }
    
    void countForallAbstract ()
    {
      if (forallAbstractWatch == null) forallAbstractWatch = new StopWatch ();
      else forallAbstractWatch.resume ();
      forallAbstractCount++;
    }
    void endCountForallAbstract ()
    {
      forallAbstractWatch.pause ();
    }
    
    void countAnd ()
    {
      if (andWatch == null) andWatch = new StopWatch ();
      else andWatch.resume ();
      
      andCount++;
    }
    void endCountAnd ()
    {
      andWatch.pause ();
    }
    void countOr ()
    {
      if (orWatch == null) orWatch = new StopWatch ();
      else orWatch.resume ();
      orCount++;
    }
    void endCountOr ()
    {
      orWatch.pause ();
    }
    
    void countNeg ()
    {
      if (negWatch == null) negWatch = new StopWatch ();
      else negWatch.resume ();
      negCount++;
    }
    void endCountNeg ()
    {
      negWatch.pause ();
    }
    
    void countOther ()
    {
      otherCount++;
    }
    
    
    public String toString ()
    {
      StringBuffer sb = new StringBuffer ();
      sb.append ("existAbstractCount: ");
      sb.append (existAbstractCount);
      sb.append (" time: " + existAbstractWatch);
      sb.append ("\nforallAbstractCount: ");
      sb.append (forallAbstractCount);
      sb.append (" time: " + forallAbstractWatch);
      sb.append ("\nandCount: ");
      sb.append (andCount);
      sb.append (" time: " + andWatch);
      sb.append ("\norCount: ");
      sb.append (orCount);
      sb.append (" time: " + orWatch);
      sb.append ("\nnegCount: ");
      sb.append (negCount);
      sb.append (" time: " + negWatch);
      sb.append ("\notherCount: ");
      sb.append (otherCount);
      sb.append ('\n');
      sb.append (mddManager.toString ());
      return sb.toString ();
    }
    
    /***
     *** Given a value in L creates a function 
     *** f(x_0, x_1, ...) = value
     ***/
    public MvSet createConstant (int value)
    {
      return createMvSet (mddManager.getConstant (value));
    }    
    
    /***
     *** Given an argument index (0 <= argIdx < n)
     *** constructs a projection function
     *** f(x_0, x_1, ...) = x_argIdx
     ***/
    public MvSet createProjection (int argIdx)
    {
      return createMvSet (mddManager.getOneVar (argIdx));
    }
    

    
    /***
     *** Creates a point function 
     *** f (args) = value
     ***          = 0 otherwise
     ***/
    public MvSet createPoint (int[] args, int value)
    {
      MDDNode mddNode = mddManager.getOne ();
      //MDDNode mddNode = mddManager.getConstant (value);

      for (int i = 0; i < args.length; i++) {
	if (args [i] == MvSetFactory.DONT_CARE ||
	    args [i] == Algebra.NO_VALUE) continue;

	mddNode = mddManager.Apply (mddNode, 
				    mddManager.getVarEquals (i, args [i]),
				    FiniteLattice.MEET);
	      
      }
      mddNode = mddManager.setEqualTo (mddNode, value);
      return createMvSet (mddNode);
    }

    public MvSet createPoint (MvState state, int value)
    {
      MDDNode mddNode = mddManager.getOne ();
      //MDDNode mddNode = mddManager.getConstant (value);
      
//        for (int i = state.getFirstVar (); i <= state.getLastVar (); i++)
//  	if (state.isSet (i))
//  	  mddNode = mddManager.Apply 
//  	    (mddNode, 
//  	     mddManager.getVarEquals (i, state.getValue (i)),
//  	     FiniteLattice.MEET);
      
      for (int i = state.getLastVar (); i >= state.getFirstVar (); i--)
	if (state.isSet (i))
	  mddNode = mddManager.Apply 
	    (mddNode, 
	     mddManager.getVarEquals (i, state.getValue (i)),
	     FiniteLattice.MEET);
      
      mddNode = mddManager.setEqualTo (mddNode, value);
      return createMvSet (mddNode);

    }

    public MvSet var (int argIdx, int argVal, int value)
    {
      //MDDNode mddNode = mddManager.getConstant (value);
      MDDNode mddNode = mddManager.getOne ();
      
      mddNode = mddManager.Apply (mddNode, 
				  mddManager.getVarEquals (argIdx, argVal),
				  FiniteLattice.MEET);
      mddManager.setEqualTo (mddNode, value);
      return createMvSet (mddNode);
    }
    

    public MvSet createCase (int argIdx, MvSet[] children)
    {
      MDDNode[] mddChildren = new MDDNode [children.length];
      for (int i = 0; i < mddChildren.length; i++)
	mddChildren [i] = ((MDDMvSet)children [i]).getMddNode ();

      return createMvSet (mddManager.makeUnique (argIdx, mddChildren));
      
    }
    

    public MvSet createMvSet (MDDNode mddNode)
    {
      return new MDDMvSet (this, mddNode);
    }

    public MDDManager getMddManager ()
    {
      return mddManager;
    }
    
    protected int getArrity ()
    {
      return arrity;
    }
    
    public FiniteLattice getLattice ()
    {
      return (FiniteLattice) mddManager.getAlgebra ();
    }
    
  }

}
