package edu.toronto.cs.mvset;

import edu.toronto.cs.algebra.*;
import edu.toronto.cs.modelchecker.PropOrderIf;


/****
 **** This is a generic interface that represents an mvset.
 **** We treat an mvset as a function from L^n -> L
 **** where L is a set (possibly a lattice) and 
 **** n is the arrity of the function
 ****/

public interface MvSet
{

  /*** 
   *** Creates a pointwise composition
   *** op is an operator: L x L -> L
   *** f is the current mv-set
   *** result h (x) = f (x) op g (x)
   ***/
  MvSet ptwiseCompose (int op, MvSet g);

  /***
   *** Creates a ptwise negation
   *** h (x) = \neg f (x)
   ***/
  MvSet ptwiseNeg ();

  /***
   *** Ptwise compare: <=, >=, =.
   ***/
  MvSet ptwiseCompare (int op, MvSet g);

  /***
   *** Restricts an argument
   *** h (x_0, x_1, x_2, ...) = f (x_0, ..., value, ...)
   *** where value is substituted at argIdx
   ***/
  MvSet restrict (int argIdx, int value);

  /***
   *** Restricts the MvSet to a particular state (or sets of states if 
   *** 'state' is partial)
   ***/
  MvSet restrict (MvState state);


  MvSet existAbstract (int[] varIndex);
  MvSet existAbstract (int var);
  MvSet existAbstract (MvSet cube);

  MvSet forallAbstract (int[] varIndex);
  MvSet forallAbstract (int var);
  MvSet forallAbstract (MvSet cube);

  /***
   *** Renames the arguments. newArgs is a map from old args to new so that
   *** h (x) = f (newArgs [0], newArgs [1], ...)
   ***/
  MvSet renameArgs (int[] newArgs);
  
  /***
   *** Evaluates this function on an input
   *** result = f (values [0], values [1], ...)
   ***/
  int evaluate (int[] values);

  /***
   *** Evaluates this function with respect to a state
   *** Note that we have a small discreptancy between what we mean 
   *** by states since here a state may contain primed and unprimed variables
   *** and thus represent a transition rather than a state
   ***/
  int evaluate (MvState state);


  /*** arrity of this function */
  int getArrity ();

  /*** gets the factory */
  MvSetFactory getFactory ();

  /*** gets the lattice for the MvSet */
  FiniteLattice getLattice ();

  void reorder ();
  /*** Returns number of nodes */
  int size ();

  /*** Utility functions */
  MvSet and (MvSet v);
  MvSet or (MvSet v);
  MvSet not ();
  MvSet leq (MvSet v);
  MvSet geq (MvSet v);
  MvSet eq (MvSet v);

  void resetStats ();

  void writeToDaVinci (String filename, PropOrderIf propOrder);

}


