package edu.toronto.cs.mvset;

import java.util.*;
import java.io.*;
import java.lang.reflect.Array;

import edu.toronto.cs.util.Logger;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.xkripke.*;
import edu.toronto.cs.xkripke.XKripke.*;
import edu.toronto.cs.modelchecker.*;

import edu.toronto.cs.mvset.*;

//import edu.toronto.cs.

public class MvState implements Cloneable
{
  int[] restrictions;
  Algebra lattice;
  int numVars;
  

  static Logger log = new Logger(MvState.class.getName());

  public MvState (int[] _restrictions, Algebra _lattice)
  {
    restrictions = _restrictions;
    numVars = restrictions.length;
    lattice = _lattice;
  }

  public Algebra getLattice ()
  {
    return lattice;
  }
  

  public MvState (int _numVars, Algebra _lattice)
  {
    lattice = _lattice;
    numVars = _numVars;
    restrictions = new int [numVars];
    Arrays.fill (restrictions, Algebra.NO_VALUE);
  }

  public Object clone ()
  {
    return new MvState ((int[])restrictions.clone (), lattice);
  }

  public MvState (XKripke.XKripkeState xstate, PropOrderIf props)
  {
    log.debug ("Initialising MvState from XKripkeState");
    //propOrder = props;
    lattice = xstate.getXKripke ().getLattice (); //xkripke.getLattice ()

    String[] propNames = xstate.getPropNames ();

    // XXX CDTW Not nice, as it's purely dependant on the fact that
    // our implementation only ever uses the numbers between 0 and
    // length * 2 for indexing.
    restrictions = new int [props.size () * 2]; // double it.
    numVars = restrictions.length;
    log.debug ("\tnumVars = " + numVars);
    Arrays.fill (restrictions, Algebra.NO_VALUE);
    
    for (int i = 0; i < propNames.length; i++)
      {
	XKripkeProp prop = xstate.getProp (propNames [i]);
	int varIndex = props.getPropOrder (prop.getName ());
	int varValue = 
	  ((FiniteLattice)lattice).getElementIdxByKey (prop.getValue ());
	log.debug ("\t  Setting prop " + xstate.getProp (propNames [i]) );
	log.debug ("\t      index of " + varIndex);
	log.debug ("\t  to Value " + prop.getValue ());
	log.debug ("\t     index " + varValue);
	
	setValue (varIndex, varValue);
      }
    log.debug ("  created: " + this.toString ());
  }
  

  public int getValue (int var)
  {
    if (inBounds (var))
      return restrictions [var];
    return Algebra.NO_VALUE;
  }

  protected boolean inBounds (int var)
  {
    return var >= getFirstVar () && var <= getLastVar ();
  }

  public int size ()
  {
    return numVars;
  }

  public int getFirstVar ()
  {
    return 0;
  }
  public int getLastVar ()
  {
    return numVars - 1;
  }

  public int numUPP (PropOrderIf props)
  {
    return props.getUnPrimedProps ().length;
  }

  public int getVarRestriction (int var, PropOrderIf props)
  {
    int [] upp = props.getUnPrimedProps ();
    return restrictions [upp [var]];
  }

  public void setValue (int var, int value)
  {
    if (inBounds (var))
      restrictions [var] = value;
  }
  public boolean isSet (int var)
  {
    return getValue (var) != Algebra.NO_VALUE;
  }
  
  public boolean isTotal ()
  {
    for (int i = 0; i < numVars; i++)
      if (restrictions [i] == Algebra.NO_VALUE)
	  return false;
    return true;
  }
  public boolean isPartial ()
  {
    return !isTotal ();
  }
  
  public String toString ()
  {
    StringBuffer sb = new StringBuffer ();
    sb.append ('{');
    for (int i = 0; i < restrictions.length; i++)
      {
	if (restrictions[i] != Algebra.NO_VALUE)
	  sb.append (lattice.getElementKeyByIdx (restrictions [i]));
	else
	  sb.append ("-");
	if (i + 1 < restrictions.length)
	  sb.append (", ");
      }
    sb.append ('}');
    return sb.toString ();	  
  }

  public int[] getRestrictions ()
  {
    return restrictions;
  }
  
  public MvSet toMvSet (MvSetFactory factory)
  {
    return factory.createPoint (this, lattice.idtimesIdx ());
  }

  public String getPropString (int i, PropOrderIf props)
  {
    StringBuffer sb = new StringBuffer ();
    int [] upp = props.getUnPrimedProps ();
    // XXX grumble i*2 not a good idea
    sb.append (props.getPropName (upp [i]));
    sb.append (" = ");
    if (restrictions[upp[i]] != Algebra.NO_VALUE)
      sb.append (lattice.getElementKeyByIdx (restrictions [upp[i]]));
    else
      sb.append ("NO_VAL");
    return sb.toString ();
  }
  
  
  public String toString (PropOrderIf props)
  {
    StringBuffer sb = new StringBuffer ();
    //sb.append ('{');
    int [] upp = props.getUnPrimedProps ();
    
    for (int i=0; i < upp.length; i++)
      {
	if (props.getPropName (upp[i]).charAt (0) == '_')
	  {
	    if (restrictions[upp[i]] == lattice.idtimesIdx ())
	      {
		sb.append (props.getPropName (upp[i]));
		sb.append (" = ");
		if (restrictions[upp[i]] != Algebra.NO_VALUE)
		  sb.append (lattice.getElementKeyByIdx (restrictions [upp[i]]));
		else
		  sb.append ("NO_VAL");
		if (i + 1 < props.size ()/2)
		sb.append (",\n");
	      }
	  }
	else
	  {
	    sb.append (props.getPropName (upp[i]));
	    sb.append (" = ");
	    if (restrictions[upp[i]] != Algebra.NO_VALUE)
	      sb.append (lattice.getElementKeyByIdx (restrictions [upp[i]]));
	    else
	      sb.append ("NO_VAL");
	    if (i + 1 < props.size ()/2)
	      sb.append (",\n");
	  }
      }
    //sb.append ('}');
    return sb.toString ();
  }  
  
}

  












