package edu.toronto.cs.tlq;

import edu.toronto.cs.algebra.*;
import edu.toronto.cs.util.*;
import java.util.*;


/**
 * class <code>UpSetLattice</code> represents a large lattice of upsets
 * Each element is encoded as an ordered list of integers, each integer
 * corresponding to a minimal element of some lattice.
 *
 * @author <a href="mailto:arie@cs.toronto.edu">Arie Gurfinkel</a>
 * @version 1.0
 */
public class UpSetLattice extends FiniteLattice //implements Algebra
{
  int mintermSize;
  
  // -- top and bottom element indecies
  int botIdx;
  int topIdx;

  // -- index of lattice elements
  List latElements = new ArrayList ();
  UniqueTable uniqueTable = new UniqueTable ();
  
  // -- need a unique table to convert between lattice elements
  // -- and their index in latElements
  

  /**
   *  <code>UpSetLattice</code> 
   * Constructs an UpSet lattice
   *
   */
  public UpSetLattice ()
  {
    this(0);
  }

  /**
   *  <code>UpSetLattice</code> 
   * Constructs an UpSet lattice with _mintermSize minterm elements
   *
   * @param _mintermSize an <code>int</code> value
   */
  public UpSetLattice (int _mintermSize)
  {
    mintermSize = _mintermSize;

    botIdx = getIdx (new int [0]);
    topIdx = getIdx (new int []{0});
    
  }

  // -- returns an upset corresponding to the minterm number 'a'
  public int getUpMinTerm (int a)
  {
    Assert.assert (a <= 32, "Currently we only support up to 32 mintrems");
    return getUpSetElement (1 << a);
  }

  public int getUpSetElement (int a)
  {
    return getIdx (new int[]{a});
  }
  
		   
  public int binOp (int l, int r, int opcode)
  {
    return binOp ((int[])latElements.get (l),
		  (int[])latElements.get (r),
		  opcode);
  }
  
  public int binOp (int[] l, int [] r, int opcode)
  {
    switch (opcode)
      {
      case FiniteLattice.MEET:
	return times (l, r);
      case FiniteLattice.JOIN:
	return plus (l, r);
      default:
	Assert.assert ("Unknown opcode: " + opcode);
      }
    return NO_VALUE;
  }
  
  public int unOp (int l, int opcode)
  {
    if (opcode == FiniteLattice.NEG)
      return neg (l);
    Assert.assert ("Unknown opcode: " + opcode);
    return NO_VALUE;
  }

  public int neg (int l)
  {
    if (l == topIdx) return botIdx;
    if (l == botIdx) return topIdx;
    Assert.assert ("Negation is only supported for top and bot elements");
    return NO_VALUE;
  }
  
  
  public int plus (int l, int r) 
  {
    if (l == botIdx) return r;
    if (r == botIdx) return l;
    if (l == topIdx || r == topIdx) return topIdx;
    
    return plus ((int[])latElements.get (l),
		 (int[])latElements.get (r));
  }
  public int plus (int[] l, int[] r)
  {
    return min (l, r);
  }
  
  
  public int idplusIdx ()
  {
    return botIdx;
  }
  
  public int times (int l, int r)
  {
    if (l == topIdx) return r;
    if (r == topIdx) return l;
    if (l == botIdx || r == botIdx) return botIdx;
    
    return times ((int[])latElements.get (l),
		  (int[])latElements.get (r));
  }
  
  public int times (int[] l, int[] r)
  {
    int[] product = new int [l.length * r.length];
    
    // -- XXX can we do better?
    for (int i = 0; i < l.length; i++)
      for (int j = 0; j < r.length; j++)
	product [(i * r.length) + j] = latJoin (l [i], r [j]);
    return min (sort (product));
  }
  
  public int idtimesIdx ()
  {
    return topIdx;
  }
  
  
  public boolean eq (int a, int b)
  {
    // -- by uniqueness of the index on lattice elements
    return a == b;
  }
  
  public boolean leq (int a, int b)
  {
    return eq (a, b) || leq ((int[])latElements.get (a),
			     (int[])latElements.get (b));
  }

  /* This one is known to work, but was replaced by a new optimized 
   * version that is not known to work in general
   */
  public boolean leq_working (int[] a, int[] b)
  {
    // -- a <= b iff for every a[i] there exits b[j] s.t. a[i] >= b[j]

    for (int i = 0; i < a.length; i++)
      {
	boolean found = false;
	for (int j = 0; j < b.length; j++)
	  if (latGeq (a [i], b [j]))
	    {
	      // -- found our b[j], no reason to search further
	      found = true;
	      break;
	    }
	// -- if found is false then we found an element a[i] which is not
	// -- above any element b[j]
	if (!found) return false;
      }
    

    return true;
  }


  /***
   *** if a[i] | b[j] == a [i]
   *** then every bit that a has off, b has off as well
   *** thus if b has a bit on, then a has the same bit on
   *** that means that (a | b = a) -> a >= b
   *** and so if a < b -> (a | b != a)
   ***/

  public boolean leq (int[] a, int[] b)
  {
    // -- a <= b iff for every a[i] there exits b[j] s.t. a[i] >= b[j]

    for (int i = 0; i < a.length; i++)
      {
	boolean found = false;
	for (int j = 0; j < b.length; j++)
	  if ((a [i] | b [j]) == a [i])
	    {
	      // -- found our b[j], no reason to search further
	      found = true;
	      break;
	    }
	// -- see the comment above why this will work
	  else if (a [i] < b [j]) break;
	
	// -- if found is false then we found an element a[i] which is not
	// -- above any element b[j]
	if (!found) return false;
      }
    

    return true;
  }



  
  public boolean geq (int a, int b)
  {
    return leq (b, a);
  }

  public int size ()
  {
    // XXX This is wrong, this is bad, but it will make MDDMgr happy
    return 2;
  }
  
  public Object getElementKeyByIdx (int idx)
  {
    // -- treat top and bot elements separatelly
    if (idx == topIdx) return "T";
    if (idx == botIdx) return "F";
    
    return latElements.get (idx);
  }
 
 
  public int getElementIdxByKey (Object key)
  {
    //    System.out.println ("Trying out a key: " + key);
    // -- top and bot are special
    if (key.equals ("T"))
      return topIdx;
    if (key.equals ("F"))
      return botIdx;
    
    if (key instanceof int[])
      return getIdx ((int[]) key);
    else 
      return NO_VALUE;
  }

  /*** Internal use ***/
  int getIdx (int[] v)
  {
    return getIdx (v, v.length);
  }
  
  int getIdx (int[] v, int len)
  {
    // -- this will check if an element with this key is already known
    // -- and return its index
    // -- if this is the first occurence, this element will be assigned
    // -- a new index, which is then returned
    return uniqueTable.getIdx (v, len);
  }
  
  

  static int[] sort (int[] a)
  {
    Arrays.sort (a);
    return a;
  }
  
  // -- our own sort routine, to be replaced by the best sort for the job
  static int[] sort (int[] a, int fromIdx, int toIdx)
  {
    // -- this is quicksort which may not be the best choice if most
    // -- of our elements are already sorted
    Arrays.sort (a, fromIdx, toIdx);
    return a;
  }
  

  // -- we don't represent the underlying lattice directly, so we need
  // -- some helper methods for it
  static int latMeet (int a, int b)
  {
    return a & b;
  }
  static int latJoin (int a, int b)
  {
    return a | b;
  }
  
  static boolean latEq (int a, int b)
  {
    return a == b;
  }
  
  static boolean latLeq (int a, int b)
  {
    // -- a <= b iff a /\ b = a
    return (a & b) == a;
  }
  
  static boolean latGeq (int a, int b)
  {
    // -- a >= b iff a \/ b = a
    return (a | b) == a;
  }
  

  // basic optimization equation
  // a < b -> !latGeq (a, b)
  // latGeq (a, b) -> a >= b

  // a > b -> !latLeq (a, b)

  int min (int[] data)
  {

    /*** check if we can quickly bail out */
    // -- 1) min of an empty array, is empty array, which is bot
    if (data.length == 0) return botIdx;
    // -- 2) min of an array of one element is just this array,
    // --    moreover if the element is 0, then this is top
    if (data.length == 1)
      return data [0] == 0 ? topIdx : getIdx (data, 1);

    /*** no we have to do the work */

    // -- assume that data is sorted
    int[] result = new int [data.length];
    int len = 1;

    // -- first element always gets copied
    result [0] = data [0];

    // now for the rest elements
    for (int i = 1; i < data.length; i++)
      {
	// -- we need to decide if we have to copy data [i] now
	// -- we don't copy it only if we have something in our
	// -- result which is below it
	boolean remove = false;
	for (int j = 0; j < len; j++)
	  // -- found something below data [i], so skip data [i]
	  if (latLeq (result [j], data [i]))
	    {
	      remove = true;
	      break;
	    }
	// copy the value if it is not removed
	if (!remove)
	  result [len++] = data [i];
      }

//     System.out.println ("Min of " + ArrayUtil.toString (data) + 
// 			" is " + ArrayUtil.toString (result, len));

    // -- our final result is sorted already so no need to resort
    return getIdx (result, len);
  }
  

  // -- min of two sorted arrays, this is basically min of a single sorted
  // -- array together with a merge sort
  int min (int[] data1, int[] data2)
  {

    // -- check if we have a simple case where no work is needed
    if (data1.length == 0 && data2.length == 0) 
      return botIdx;
    else if (data1.length == 0)
      return getIdx (data2, data2.length);
    else if (data2.length == 0)
      return getIdx (data1, data1.length);
    
    
    // -- no, we have to do the work
    int[] result = new int [data1.length + data2.length];
    int len = 0;
    
    int pos1 = 0;
    int pos2 = 0;
    

    // -- while there are unprocessed elements
    while (pos1 < data1.length || pos2 < data2.length)
      {
	int element;
    
	// -- pick the smallest element between data1 and data2
	// -- if either data1 or data2 is exhausted then trivial
	// -- otherwise pick the smallest between the current elements
	if (pos1 == data1.length)
	  element = data2 [pos2++];
	else if (pos2 == data2.length)
	  element = data1 [pos1++];
	else if (data1 [pos1] == data2 [pos2])
	  {
	    // -- if an element belongs to both data1 and data2 
	    // -- use it from data1 and ignore data2.
	    // -- more over in this case we don't need to check 
	    // -- if the element is irredundant, it can be proven that 
	    // -- it is
	    result [len++] = data1 [pos1++];
	    pos2++;
	    continue;
	  }
	else 
	  element = 
	    data1 [pos1] < data2 [pos2] ? data1 [pos1++] : data2 [pos2++];
    
	boolean remove = false;
	// -- check if this element should be part of the irreduntant set
	for (int i = 0; i < len; i++)
	  if (latLeq (result [i], element))
	    {
	      remove = true;
	      break;
	    }
    
	if (!remove) result [len++] = element;
      }

    //     System.out.println ("Min of " + ArrayUtil.toString (data1) + 
    // 			" and " + ArrayUtil.toString (data2) + 
    // 			" is " + ArrayUtil.toString (result, len));


    return getIdx (result, len);
  }
  

  // -- find all minimal elements in the set
  int min_unsorted (int[] data)
  {

//     System.out.println ("min of " + 
// 			arrayToString (data));

    // -- this will hold a result, its size is at most data.length
    int[] result = new int [data.length];
    // -- counts how many elements we actually put in result
    int len = 0;
    


    // XXX This is not the best possible algorithm! FIX IT
    for (int i = 0; i < data.length; i++)
      {
	boolean remove = false;
	// -- consider element data [i]
	for (int j = 0; j < data.length; j++)
	  {
	    // -- check if data [i] should be removed from the min
	    // -- two cases
	    // -- 1) data [i] == data [j] and i > j -- in this case 
	    //       data [j] stays and data [i] is removed
	    // -- 2) data [i] > data [j]
	    if (latEq (data [i], data [j]))
	      remove = i > j;
	    else 
	      remove = latGeq (data [i], data [j]);
	    
	    // -- stop as soon as we've made our decision
	    if (remove) break;
	      
	  }
	if (!remove) result [len++] = data [i];
      }


    
    // -- return unique index as the result
    return getIdx (sort (result, 0, len), len);
  }
  
  int min_unsorted (int[] data1, int [] data2)
  {
    // XXX just for now
    int[] data = new int [data1.length + data2.length];
    System.arraycopy (data1, 0, data, 0, data1.length);
    System.arraycopy (data2, 0, data, data1.length, data2.length);
    return min (data);
  }
  

  public BitSet getIrredundantJoin (BitSet values)
  {
    // XXX yet another n^2 algorithm

    // --  clone since we are doing a destructive operation
    BitSet result = (BitSet)values.clone ();
    
    int bitLength = values.length ();
    for (int i = 0; i < bitLength; i++)
      {
	// -- skip over empty bits
	if (!values.get (i)) continue;

	// -- got a set bit, compare it to the rest
	for (int j = 0; j < result.length (); j++)
	  // -- if bit 'j' is set, and current value is less then j
	  // -- then current value is redundant and should be removed
	  if (i != j && result.get (j) && leq (i, j)) result.clear (i);
      }
    return result;
  }

  public BitSet getIrredundantMeet (BitSet values)
  {
    // XXX yet another n^2 algorithm

    // --  clone since we are doing a destructive operation
    BitSet result = (BitSet)values.clone ();
    
    int bitLength = values.length ();
    for (int i = 0; i < bitLength; i++)
      {
	// -- skip over empty bits
	if (!values.get (i)) continue;

	// -- got a set bit, compare it to the rest
	for (int j = 0; j < result.length (); j++)
	  // -- if bit 'j' is set, and current value is greater then j
	  // -- then current value is redundant and should be removed
	  if (i != j && result.get (j) && leq (j, i)) result.clear (i);
      }
    return result;
  }
  

  class UniqueTableElement
  {
    int[] data;
    int len;
    

    public UniqueTableElement ()
    {
      this (null, -1);
    }
    
    public UniqueTableElement (int[] _data, int _len)
    {
      setData (_data);
      setLength (_len);
    }
    
    public void setData (int[] v)
    {
      data = v;
    }
    public void setLength (int v)
    {
      len = v;
    }
    
    public boolean equals (Object o)
    {
      if (o.getClass () == UniqueTableElement.class)
	return equals ((UniqueTableElement) o);
      return false;
    }
    
    public boolean equals (UniqueTableElement element)
    {
      // -- length must be the same
      if (len != element.len) return false;
      
      for (int i = 0; i < element.len; i++)
	if (data [i] != element.data [i]) return false;
      
      return true;
    }


    // XXX This should be made uniform for MDDNodes and the rest
    public static final long[] primes = 
    {12582917L, 4256249L, 741457L, 1618033999L};

    public int hashCode ()
    {
      long hash = 0;

      int hashLen = len > primes.length ? primes.length : len;
      
      for (int i = 0; i < hashLen; i++)
	hash += data [i] * primes [i];
      
      return (int)hash;
    }
    
    public String toString ()
    {
      return ArrayUtil.toString (data, len);
    }
    
    
  }
  

  class UniqueTable 
  {
    Map map = new HashMap ();

    UniqueTableElement lookup = new UniqueTableElement ();
    
    int getIdx (int[] data, int len)
    {
      lookup.setData (data);
      lookup.setLength (len);   
      Integer idx = (Integer)map.get (lookup);
      
      if (idx != null) return idx.intValue ();
      
      // -- did not find the element, so add it
      int[] result = new int [len];
      System.arraycopy (data, 0, result, 0, len);

      idx = new Integer (latElements.size ());
      latElements.add (result);
      map.put (new UniqueTableElement (result, result.length), idx);
      return idx.intValue ();
    }
    
  }
  

  public static void main (String[] args)
  {
    UpSetLattice lat = new UpSetLattice (2);
    System.out.println ("topIdx: " + lat.topIdx);
    System.out.println ("botIdx: " + lat.botIdx);

    System.out.println ("bot: " + 
			arrayToString (lat.latElements.get (lat.botIdx)));
    System.out.println ("top: " + 
			arrayToString (lat.latElements.get (lat.topIdx)));

    int a = lat.getUpSetElement (1);
    int b = lat.getUpSetElement (2);
    int c = lat.getUpSetElement (3);
    int r = lat.plus (a, b);
    r = lat.times (r, c);
    
    System.out.println ("r: " + 
			arrayToString (lat.latElements.get (r)));

    int v = lat.getUpSetElement (5);
    System.out.println ("Asked for 5: got: " + v);
    v = lat.getUpSetElement (5);
    System.out.println ("Asked for 5: got: " + v);
  }


  static String arrayToString (Object o)
  {
    int[] a = (int[]) o;
    return ArrayUtil.toString (a);
  }

  public void dumpValueMap (java.io.PrintWriter out)
  {
    out.println ("Value map is not supported by: " + 
		 UpSetLattice.class.getName ());
  }
  
  
}
